from sismic.bdd import map_action, map_assertion


map_action('I open the door', 'I send event door_opened')
map_action('I close the door', 'I send event door_closed')
map_action('I place an item in the oven', 'I send event item_placed')
map_action('I press increase timer button {time} times', 'I repeat "I send event timer_inc" {time} times')
map_action('I press increase power button', 'I send event power_inc')
map_action('I press start button', 'I send event cooking_start')
map_action('I press stop button', 'I send event cooking_stop')
map_action('{tick} seconds elapsed', 'I repeat "I send event timer_tick" {tick} times')

map_assertion('Heating turns on', 'Event heating_on is fired')
map_assertion('Heating does not turn on', 'Event heating_on is not fired')
map_assertion('heating turns off', 'Event heating_off is fired')
map_assertion('lamp turns on', 'Event lamp_switch_on is fired')
map_assertion('lamp turns off', 'Event lamp_switch_off is fired')



from typing import List, Mapping

from .evaluator import Evaluator
from ..model import Event

__all__ = ['DummyEvaluator']


class DummyEvaluator(Evaluator):
    """
    A dummy evaluator that does nothing and evaluates every condition to True.
    """

    def __init__(self, interpreter=None, *, initial_context=None):
        super().__init__(interpreter, initial_context=initial_context)

    @property
    def context(self):
        return dict()

    def _evaluate_code(self, code: str, *, additional_context: Mapping=None) -> bool:
        return True

    def _execute_code(self, code: str, *, additional_context: Mapping=None) -> List[Event]:
        return []



import os
import shutil
import tempfile

from typing import Union, List, Callable
from behave import given, when, then
from behave.__main__ import run_behave
from behave.configuration import Configuration

from ..interpreter import Interpreter
from ..model import Statechart


__all__ = ['map_action', 'map_assertion', 'execute_bdd']


def map_action(step_text: str, existing_step_or_steps: Union[str, List[str]]) -> None:
    """
    Map new "given"/"when" steps to one or many existing one(s).
    Parameters are propagated to the original step(s) as well, as expected.

    Examples:

     - map_action('I open door', 'I send event open_door')
     - map_action('Event {name} has to be sent', 'I send event {name}')
     - map_action('I do two things', ['First thing to do', 'Second thing to do'])

    :param step_text: Text of the new step, without the "given" or "when" keyword.
    :param existing_step_or_steps: existing step, without the "given" or "when" keyword. Could be a list of steps.
    """
    if not isinstance(existing_step_or_steps, str):
        existing_step_or_steps = '\nand '.join(existing_step_or_steps)

    @given(step_text)
    def _(context, **kwargs):
        context.execute_steps('Given ' + existing_step_or_steps.format(**kwargs))

    @when(step_text)
    def _(context, **kwargs):
        context.execute_steps('When ' + existing_step_or_steps.format(**kwargs))


def map_assertion(step_text: str, existing_step_or_steps: Union[str, List[str]]) -> None:
    """
    Map a new "then" step to one or many existing one(s).
    Parameters are propagated to the original step(s) as well, as expected.

    map_assertion('door is open', 'state door open is active')
    map_assertion('{x} seconds elapsed', 'I wait for {x} seconds')
    map_assertion('assert two things', ['first thing to assert', 'second thing to assert'])

    :param step_text: Text of the new step, without the "then" keyword.
    :param existing_step_or_steps: existing step, without "then" keyword. Could be a list of steps.
    """
    if not isinstance(existing_step_or_steps, str):
        existing_step_or_steps = '\nand '.join(existing_step_or_steps)

    @then(step_text)
    def _(context, **kwargs):
        context.execute_steps('Then ' + existing_step_or_steps.format(**kwargs))


def execute_bdd(statechart: Statechart,
                feature_filepaths: List[str],
                *,
                step_filepaths: List[str]=None,
                property_statecharts: List[Statechart]=None,
                interpreter_klass: Callable[[Statechart], Interpreter]=Interpreter,
                debug_on_error: bool=False,
                behave_parameters: List[str]=None) -> int:
    """
    Execute BDD tests for a statechart.

    :param statechart: statechart to test
    :param feature_filepaths: list of filepaths to feature files.
    :param step_filepaths: list of filepaths to step definitions.
    :param property_statecharts: list of property statecharts
    :param interpreter_klass: a callable that accepts a statechart and returns an Interpreter
    :param debug_on_error: set to True to drop to (i)pdb in case of error.
    :param behave_parameters: additional CLI parameters used by Behave (see http://behave.readthedocs.io/en/latest/behave.html#command-line-arguments)
    :return: exit code of behave CLI.
    """
    # Default values
    step_filepaths = step_filepaths if step_filepaths else []
    property_statecharts = property_statecharts if property_statecharts else []
    behave_parameters = behave_parameters if behave_parameters else []

    # If debug_on_error, disable captured stdout, otherwise it hangs
    if debug_on_error and '--capture' not in behave_parameters:
        behave_parameters.append('--no-capture')

    # Create temporary directory to put everything inside
    with tempfile.TemporaryDirectory() as tempdir:
        # Create configuration for Behave
        config = Configuration(behave_parameters)

        # Paths to features
        config.paths = feature_filepaths

        # Copy environment
        with open(os.path.join(tempdir, 'environment.py'), 'w') as environment:
            environment.write('from sismic.bdd.environment import *')

        # Path to environment
        config.environment_file = os.path.join(tempdir, 'environment.py')

        # Add predefined steps
        os.mkdir(os.path.join(tempdir, 'steps'))
        with open(os.path.join(tempdir, 'steps', '__steps.py'), 'w') as step:
            step.write('from sismic.bdd.steps import *')

        # Copy provided steps, if any
        for step_filepath in step_filepaths:
            shutil.copy(step_filepath, os.path.join(tempdir, 'steps', os.path.split(step_filepath)[-1]))

        # Path to steps
        config.steps_dir = os.path.join(tempdir, 'steps')

        # Put statechart and properties in user data
        config.update_userdata({
            'statechart': statechart,
            'interpreter_klass': interpreter_klass,
            'property_statecharts': property_statecharts,
            'debug_on_error': debug_on_error,
        })

        # Run behave
        return run_behave(config)



from collections import OrderedDict
from typing import Any, List, Mapping, MutableMapping, Optional, Tuple, cast

from ..exceptions import StatechartError
from ..model import (ActionStateMixin, BasicState, CompositeStateMixin,
                     CompoundState, DeepHistoryState, FinalState,
                     OrthogonalState, ShallowHistoryState, Statechart,
                     StateMixin, Transition, TransitionStateMixin)

__all__ = ['import_from_dict', 'export_to_dict']


def import_from_dict(data: Mapping[str, Any]) -> Statechart:
    data = data['statechart']

    statechart = Statechart(name=data['name'],
                            description=data.get('description', None),
                            preamble=data.get('preamble', None))

    states = []  # (StateMixin instance, parent name)
    transitions = []  # Transition instances
    # (State dict, parent name)
    data_to_consider = [(data['root state'], None)]  # type: List[Tuple[Mapping[str, Any], Optional[str]]]

    while data_to_consider:
        state_data, parent_name = data_to_consider.pop()

        # Get state
        try:
            state = _import_state_from_dict(state_data)
        except StatechartError:
            raise
        except Exception as e:
            raise StatechartError('Unable to load given YAML') from e
        states.append((state, parent_name))

        # Get substates
        if isinstance(state, CompoundState):
            for substate_data in state_data['states']:
                data_to_consider.append((substate_data, state.name))
        elif isinstance(state, OrthogonalState):
            for substate_data in state_data['parallel states']:
                data_to_consider.append((substate_data, state.name))

        # Get transition(s)
        for transition_data in state_data.get('transitions', []):
            try:
                transition = _import_transition_from_dict(state.name, transition_data)
            except StatechartError:
                raise
            except Exception as e:
                raise StatechartError('Unable to load given YAML') from e
            transitions.append(transition)

    # Register on statechart
    for state, parent in states:
        statechart.add_state(state, parent)
    for transition in transitions:
        statechart.add_transition(transition)

    return statechart


def _import_transition_from_dict(state_name: str, transition_d: Mapping[str, Any]) -> Transition:
    """
    Return a Transition instance from given dict.

    :param state_name: name of the state in which the transition is defined
    :param transition_d: a dictionary containing transition data
    :return: an instance of Transition
    """
    event = transition_d.get('event', None)
    guard = transition_d.get('guard', None)
    action = transition_d.get('action', None)

    transition = Transition(
        state_name,
        transition_d.get('target', None),
        event.strip() if event else None,
        guard.strip() if guard else None,
        action.strip() if action else None,
    )

    # Preconditions, postconditions and invariants
    for condition in transition_d.get('contract', []):
        if condition.get('before', None):
            transition.preconditions.append(condition['before'].strip())
        elif condition.get('after', None):
            transition.postconditions.append(condition['after'].strip())
        elif condition.get('always', None):
            transition.invariants.append(condition['always'].strip())

    return transition


def _import_state_from_dict(state_d: Mapping[str, Any]) -> StateMixin:
    """
    Return the appropriate type of state from given dict.

    :param state_d: a dictionary containing state data
    :return: a specialized instance of State
    """
    name = state_d['name']  # type: str
    stype = state_d.get('type', None)

    on_entry = state_d.get('on entry', None)  # type: Optional[str]
    on_entry = on_entry.strip() if on_entry else None
    on_exit = state_d.get('on exit', None)  # type: Optional[str]
    on_exit = on_exit.strip() if on_exit else None

    state = None  # type: Any

    if stype == 'final':
        state = FinalState(name, on_entry=on_entry, on_exit=on_exit)
    elif stype == 'shallow history':
        state = ShallowHistoryState(name, on_entry=on_entry, on_exit=on_exit, memory=state_d.get('memory', None))
    elif stype == 'deep history':
        state = DeepHistoryState(name, on_entry=on_entry, on_exit=on_exit, memory=state_d.get('memory', None))
    elif stype is None:
        substates = state_d.get('states', None)
        parallel_substates = state_d.get('parallel states', None)

        if substates and parallel_substates:
            raise StatechartError('{} cannot declare both a "states" and a "parallel states" property'.format(name))
        elif substates:
            state = CompoundState(name, initial=state_d.get('initial', None), on_entry=on_entry, on_exit=on_exit)
        elif parallel_substates:
            state = OrthogonalState(name, on_entry=on_entry, on_exit=on_exit)
        else:
            state = BasicState(name, on_entry=on_entry, on_exit=on_exit)
    else:
        raise StatechartError('Unknown type {} for state {}'.format(stype, name))

    # Preconditions, postconditions and invariants
    for condition in state_d.get('contract', []):
        if condition.get('before', None):
            state.preconditions.append(condition['before'].strip())
        elif condition.get('after', None):
            state.postconditions.append(condition['after'].strip())
        elif condition.get('always', None):
            state.invariants.append(condition['always'].strip())
    return state


def export_to_dict(statechart: Statechart, ordered=True) -> Mapping[str, Any]:
    """
    Export given StateChart instance to a dict.

    :param statechart: a StateChart instance
    :param ordered: set to True to use an ordereddict instead of a dict
    :return: a dict that can be used in *_import_from_dict*
    """
    d = OrderedDict() if ordered else {}  # type: MutableMapping
    d['name'] = statechart.name
    if statechart.description:
        d['description'] = statechart.description
    if statechart.preamble:
        d['preamble'] = statechart.preamble

    d['root state'] = _export_state_to_dict(statechart, statechart.root, ordered)

    return {'statechart': d}


def _export_state_to_dict(statechart: Statechart, state_name: str, ordered=True) -> Mapping[str, Any]:
    data = OrderedDict() if ordered else {}

    state = statechart.state_for(state_name)

    data['name'] = state.name
    if isinstance(state, ShallowHistoryState):
        data['type'] = 'shallow history'
        if state.memory:
            data['memory'] = state.memory
    elif isinstance(state, DeepHistoryState):
        data['type'] = 'deep history'
        if state.memory:
            data['memory'] = state.memory
    elif isinstance(state, FinalState):
        data['type'] = 'final'

    if isinstance(state, ActionStateMixin):
        if state.on_entry:
            data['on entry'] = state.on_entry
        if state.on_exit:
            data['on exit'] = state.on_exit

    if isinstance(state, CompoundState):
        if state.initial:
            data['initial'] = state.initial

    preconditions = getattr(state, 'preconditions', [])
    postconditions = getattr(state, 'postconditions', [])
    invariants = getattr(state, 'invariants', [])
    if preconditions or postconditions or invariants:
        conditions = []
        for condition in preconditions:
            conditions.append({'before': condition})
        for condition in postconditions:
            conditions.append({'after': condition})
        for condition in invariants:
            conditions.append({'always': condition})
        data['contract'] = conditions

    if isinstance(state, TransitionStateMixin):
        # event, guard, target, action
        transitions = statechart.transitions_from(cast(StateMixin, state).name)
        if len(transitions) > 0:
            data['transitions'] = []

            for transition in transitions:
                transition_data = OrderedDict() if ordered else {}
                if transition.event:
                    transition_data['event'] = transition.event
                if transition.guard:
                    transition_data['guard'] = transition.guard
                if transition.target:
                    transition_data['target'] = transition.target
                if transition.action:
                    transition_data['action'] = transition.action

                preconditions = getattr(transition, 'preconditions', [])
                postconditions = getattr(transition, 'postconditions', [])
                invariants = getattr(transition, 'invariants', [])
                if preconditions or postconditions or invariants:
                    conditions = []
                    for condition in preconditions:
                        conditions.append({'before': condition})
                    for condition in postconditions:
                        conditions.append({'after': condition})
                    for condition in invariants:
                        conditions.append({'always': condition})
                    transition_data['contract'] = conditions

                data['transitions'].append(transition_data)

    if isinstance(state, CompositeStateMixin):
        children = statechart.children_for(cast(StateMixin, state).name)
        children_data = [_export_state_to_dict(statechart, child, ordered) for child in children]

        if isinstance(state, CompoundState):
            data['states'] = children_data
        elif isinstance(state, OrthogonalState):
            data['parallel states'] = children_data

    return data



from typing import Any

__all__ = ['Event', 'InternalEvent', 'MetaEvent']


class Event:
    """
    Simple event with a name and (optionally) some data.
    Unless the attribute already exists, each key from *data* is exposed as an attribute
    of this class.

    The list of defined attributes can be obtained using *dir(event)*.

    :param name: Name of the event
    :param data: additional data (mapping, dict-like)
    """

    __slots__ = ['name', 'data']

    def __init__(self, name: str, **additional_parameters: Any) -> None:
        self.name = name
        self.data = additional_parameters

    def __eq__(self, other):
        return (self.name == other.name and self.data == other.data) if isinstance(other, Event) else NotImplemented

    def __getattr__(self, attr):
        try:
            return self.data[attr]
        except KeyError:
            raise AttributeError('{} has no attribute {}'.format(self, attr))

    def __getstate__(self):
        # For pickle and implicitly for multiprocessing
        return self.name, self.data

    def __setstate__(self, state):
        # For pickle and implicitly for multiprocessing
        self.name, self.data = state

    def __hash__(self):
        return hash(self.name)

    def __dir__(self):
        return ['name'] + list(self.data.keys())

    def __repr__(self):
        if self.data:
            return '{}({!r}, {})'.format(self.__class__.__name__,
                                         self.name,
                                         ', '.join('{}={!r}'.format(k, v) for k, v in self.data.items()))
        else:
            return '{}({!r})'.format(self.__class__.__name__, self.name)


class InternalEvent(Event):
    """
    Subclass of Event that represents an internal event.
    """
    pass


class MetaEvent(Event):
    """
    Subclass of Event that represents a MetaEvent, as used in property statecharts.
    """
    pass


import collections
import copy
from functools import partial
from types import CodeType
from typing import Any, Callable, Dict, Iterator, List, Mapping, Optional

from . import Evaluator
from ..exceptions import CodeEvaluationError
from ..model import Statechart, StateMixin, Transition, Event, InternalEvent, MetaEvent

__all__ = ['PythonEvaluator']


class FrozenContext(collections.Mapping):
    """
    A shallow copy of a context. The keys of the underlying context are
    exposed as attributes.
    """
    __slots__ = ['__frozencontext']

    def __init__(self, context: Mapping) -> None:
        self.__frozencontext = {k: copy.copy(v) for k, v in context.items()}

    def __getattr__(self, item):
        try:
            return self.__frozencontext[item]
        except KeyError:
            raise AttributeError('{} has no attribute {}'.format(self, item))

    def __getstate__(self):
        return self.__frozencontext

    def __setstate__(self, state):
        self.__frozencontext = state

    def __getitem__(self, key):
        return self.__frozencontext[key]

    def __len__(self):
        return len(self.__frozencontext)

    def __iter__(self):
        return iter(self.__frozencontext)


def create_send_function(event_list: List[Event], event_klass: Callable[..., Event]) -> Callable[..., None]:
    """
    Create and return a callable that takes a name and additional parameters, builds an Event of type event_klass,
    and add it into given *event_list*.

    :param event_list: list to complement
    :param event_klass: Event to create inside the function
    :return: the newly created function
    """

    def send(name, **kwargs):
        event = event_klass(name, **kwargs)
        event_list.append(event)

    return send


class PythonEvaluator(Evaluator):
    """
    A code evaluator that understands Python.

    Depending on the method that is called, the context can expose additional values:

    - On both code execution and code evaluation:
        - A *time: float* value that represents the current time exposed by the interpreter.
        - An *active(name: str) -> bool* Boolean function that takes a state name and return *True* if and only
          if this state is currently active, ie. it is in the active configuration of the ``Interpreter`` instance
          that makes use of this evaluator.
    - On code execution:
        - A *send(name: str, **kwargs) -> None* function that takes an event name and additional keyword parameters and
          raises an internal event with it.
        - A *notify(name: str, **kwargs) -> None* function that takes an event name and additional keyword parameters and
          raises a meta-event with it. Meta-events are only sent to bound property statecharts.
        - If the code is related to a transition, the *event: Event* that fires the transition is exposed.
    - On guard or contract evaluation:
        - If the code is related to a transition, the *event: Event* that fires the transition is exposed.
    - On guard or contract (except preconditions) evaluation:
        - An *after(sec: float) -> bool* Boolean function that returns *True* if and only if the source state
          was entered more than *sec* seconds ago. The time is evaluated according to Interpreter's clock.
        - A *idle(sec: float) -> bool* Boolean function that returns *True* if and only if the source state
          did not fire a transition for more than *sec* ago. The time is evaluated according to Interpreter's clock.
    - On contract (except preconditions) evaluation:
        - A variable *__old__* that has an attribute *x* for every *x* in the context when either the state
          was entered (if the condition involves a state) or the transition was processed (if the condition
          involves a transition). The value of *__old__.x* is a shallow copy of *x* at that time.
    - On contract evaluation:
        - A *sent(name: str) -> bool* function that takes an event name and return True if an event with the same name
          was sent during the current step.
        - A *received(name: str) -> bool* function  that takes an event name and return True if an event with the
          same name is currently processed in this step.

    If an exception occurred while executing or evaluating a piece of code, it is propagated by the
    evaluator.

    :param interpreter: the interpreter that will use this evaluator,
        is expected to be an *Interpreter* instance
    :param initial_context: a dictionary that will be used as *__locals__*
    """

    def __init__(self, interpreter=None, *, initial_context: Mapping[str, Any]=None) -> None:
        super().__init__(interpreter, initial_context=initial_context)

        self._context = {} if initial_context is None else initial_context  # type: Mapping[str, Any]
        self._interpreter = interpreter

        # Memory and entry time
        self._memory = {}  # type: Dict[int, Mapping]
        self._entry_time = {}  # type: Dict[str, float]
        self._idle_time = {}  # type: Dict[str, float]

        # Precompiled code
        self._evaluable_code = {}  # type: Dict[str, CodeType]
        self._executable_code = {}  # type: Dict[str, CodeType]

        # Intercept sent and received events
        self._sent_events = []  # type: List[Event]
        if self._interpreter is not None:
            self._interpreter.bind(self._sent_events.append)
        self._received_event = None  # type: Optional[Event]

    @property
    def context(self) -> Mapping:
        return self._context

    def on_step_starts(self, event: Optional[Event]=None) -> None:
        """
        Called each time the interpreter starts a macro step.

        :param event: Optional processed event
        """
        self._sent_events.clear()
        self._received_event = event

    def _received(self, name: str) -> bool:
        """
        :param name: name of an event
        :return: True if given event name was received in current step.
        """
        return getattr(self._received_event, 'name', None) == name

    def _sent(self, name: str) -> bool:
        """
        :param name: name of an event
        :return: True if given event name was sent during this step.
        """
        return any((name == e.name for e in self._sent_events))

    def _active(self, name: str) -> bool:
        """
        Return True if given state name is active.

        :param name: name of a state
        :return: True if given state name is active.
        """
        return name in self._interpreter.configuration

    def _after(self, name: str, seconds: float) -> bool:
        """
        Return True if given state was entered more than *seconds* ago.

        :param name: name of a state
        :param seconds: elapsed time
        :return: True if given state was entered more than *seconds* ago.
        """
        return self._interpreter.time - seconds >= self._entry_time[name]

    def _idle(self, name: str, seconds: float) -> bool:
        """
        Return True if given state was the target of a transition more than *seconds* ago.

        :param name: name of a state
        :param seconds: elapsed time
        :return: True if given state was the target of a transition more than *seconds* ago.
        """
        return self._interpreter.time - seconds >= self._idle_time[name]

    def _evaluate_code(self, code: Optional[str], *, additional_context: Mapping=None) -> bool:
        """
        Evaluate given code using Python.

        :param code: code to evaluate
        :param additional_context: an optional additional context
        :return: truth value of *code*
        """
        if code is None:
            return True

        compiled_code = self._evaluable_code.get(code, None)
        if compiled_code is None:
            compiled_code = self._evaluable_code.setdefault(code, compile(code, '<string>', 'eval'))

        exposed_context = {
            'active': self._active,
            'time': self._interpreter.time,
        }
        exposed_context.update(additional_context if additional_context is not None else {})

        try:
            return eval(compiled_code, exposed_context, self._context)  # type: ignore
        except Exception as e:
            raise CodeEvaluationError('The above exception occurred while evaluating:\n{}'.format(code)) from e

    def _execute_code(self, code: Optional[str], *, additional_context: Mapping=None) -> List[Event]:
        """
        Execute given code using Python.

        :param code: code to execute
        :param additional_context: an optional additional context
        :return: a list of sent events
        """
        if code is None:
            return []

        compiled_code = self._executable_code.get(code, None)
        if compiled_code is None:
            compiled_code = self._executable_code.setdefault(code, compile(code, '<string>', 'exec'))

        # Handle sent events
        sent_events = []  # type: List[Event]

        exposed_context = {
            'active': self._active,
            'send': create_send_function(sent_events, InternalEvent),
            'notify': create_send_function(sent_events, MetaEvent),
            'time': self._interpreter.time,
        }
        exposed_context.update(additional_context if additional_context is not None else {})

        try:
            exec(compiled_code, exposed_context, self._context)  # type: ignore
            return sent_events
        except Exception as e:
            raise CodeEvaluationError('The above exception occurred while executing:\n{}'.format(code)) from e

    def execute_statechart(self, statechart: Statechart):
        """
        Execute the initial code of a statechart.
        This method is called at the very beginning of the execution.

        :param statechart: statechart to consider
        """
        if statechart.preamble:
            events = self._execute_code(statechart.preamble)
            if len(events) > 0:
                raise CodeEvaluationError('Events cannot be raised by statechart preamble')

    def evaluate_guard(self, transition: Transition, event: Optional[Event]=None) -> bool:
        """
        Evaluate the guard for given transition.

        :param transition: the considered transition
        :param event: instance of *Event* if any
        :return: truth value of *code*
        """
        additional_context = {
            'event': event,
            'after': partial(self._after, transition.source),
            'idle': partial(self._idle, transition.source),
        }
        return self._evaluate_code(getattr(transition, 'guard', None), additional_context=additional_context)

    def execute_action(self, transition: Transition, event: Optional[Event]=None) -> List[Event]:
        """
        Execute the action for given transition.
        This method is called for every transition that is processed, even those with no *action*.

        :param transition: the considered transition
        :param event: instance of *Event* if any
        :return: a list of sent events
        """
        self._idle_time[transition.source] = self._interpreter.time

        return self._execute_code(getattr(transition, 'action', None), additional_context={'event': event})

    def execute_on_entry(self, state: StateMixin) -> List[Event]:
        """
        Execute the on entry action for given state.
        This method is called for every state that is entered, even those with no *on_entry*.

        :param state: the considered state
        :return: a list of sent events
        """
        self._entry_time[state.name] = self._interpreter.time
        self._idle_time[state.name] = self._interpreter.time

        return self._execute_code(getattr(state, 'on_entry', None))

    def execute_on_exit(self, state: StateMixin) -> List[Event]:
        """
        Execute the on exit action for given state.
        This method is called for every state that is exited, even those with no *on_exit*.

        :param state: the considered state
        :return: a list of sent events
        """
        return self._execute_code(getattr(state, 'on_exit', None))

    def evaluate_preconditions(self, obj, event: Optional[Event]=None) -> Iterator[str]:
        """
        Evaluate the preconditions for given object (either a *StateMixin* or a
        *Transition*) and return a list of conditions that are not satisfied.

        :param obj: the considered state or transition
        :param event: an optional *Event* instance, in the case of a transition
        :return: list of unsatisfied conditions
        """
        state_name = obj.source if isinstance(obj, Transition) else obj.name

        additional_context = {'event': event} if isinstance(obj, Transition) else {}  # type: Dict[str, Any]
        additional_context.update({
            'received': self._received,
            'sent': self._sent
        })  # type: ignore

        # Only needed if there is an invariant, a postcondition or a sequential condition
        if len(getattr(obj, 'invariants', [])) > 0 or len(getattr(obj, 'postconditions', [])) > 0 or len(
                getattr(obj, 'sequences', [])) > 0:
            self._memory[id(obj)] = FrozenContext(self._context)

        return filter(
            lambda c: not self._evaluate_code(c, additional_context=additional_context),
            getattr(obj, 'preconditions', [])
        )

    def evaluate_invariants(self, obj, event: Optional[Event]=None) -> Iterator[str]:
        """
        Evaluate the invariants for given object (either a *StateMixin* or a
        *Transition*) and return a list of conditions that are not satisfied.

        :param obj: the considered state or transition
        :param event: an optional *Event* instance, in the case of a transition
        :return: list of unsatisfied conditions
        """
        state_name = obj.source if isinstance(obj, Transition) else obj.name

        additional_context = {'event': event} if isinstance(obj, Transition) else {}  # type: Dict[str, Any]
        additional_context.update({
            '__old__': self._memory.get(id(obj), None),
            'after': partial(self._after, state_name),
            'idle': partial(self._idle, state_name),
            'received': self._received,
            'sent': self._sent,
        })

        return filter(
            lambda c: not self._evaluate_code(c, additional_context=additional_context),
            getattr(obj, 'invariants', [])
        )

    def evaluate_postconditions(self, obj, event: Optional[Event]=None) -> Iterator[str]:
        """
        Evaluate the postconditions for given object (either a *StateMixin* or a
        *Transition*) and return a list of conditions that are not satisfied.

        :param obj: the considered state or transition
        :param event: an optional *Event* instance, in the case of a transition
        :return: list of unsatisfied conditions
        """
        state_name = obj.source if isinstance(obj, Transition) else obj.name

        additional_context = {'event': event} if isinstance(obj, Transition) else {}  # type: Dict[str, Any]
        additional_context.update({
            '__old__': self._memory.get(id(obj), None),
            'after': partial(self._after, state_name),
            'idle': partial(self._idle, state_name),
            'received': self._received,
            'sent': self._sent,
        })

        return filter(
            lambda c: not self._evaluate_code(c, additional_context=additional_context),
            getattr(obj, 'postconditions', [])
        )

    def __getstate__(self):
        attributes = self.__dict__.copy()
        attributes['_executable_code'] = dict()  # Code fragment cannot be pickled
        attributes['_evaluable_code'] = dict()  # Code fragment cannot be pickled
        return attributes



from copy import deepcopy
from typing import Callable, Dict, Iterable, List, Optional, Union, cast

from ..exceptions import StatechartError

from .elements import (CompositeStateMixin, CompoundState, HistoryStateMixin,
                       StateMixin, Transition, TransitionStateMixin)

__all__ = ['Statechart']


class Statechart:
    """
    Python structure for a statechart

    :param name: Name of this statechart
    :param description: optional description
    :param preamble: code to execute to bootstrap the statechart
    """

    def __init__(self, name: str, description: str=None, preamble: str=None) -> None:
        self.name = name
        self.description = description
        self._preamble = preamble

        self._states = {}  # type: Dict[str, StateMixin]
        self._parent = {}  # type: Dict[str, Optional[str]]
        self._children = {}  # type: Dict[Optional[str], List[str]]
        self._transitions = []  # type: List[Transition]

        self._children[None] = []  # Root state

    @property
    def root(self) -> Optional[str]:
        """
        Root state name
        """
        for name, parent in self._parent.items():
            if parent is None:
                return name
        return None

    @property
    def preamble(self):
        """
        Preamble code
        """
        return self._preamble

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self.name)

    # ######### STATES ##########

    @property
    def states(self):
        """
        List of state names in lexicographic order.
        """
        return sorted(self._states.keys())

    def state_for(self, name: str) -> StateMixin:
        """
        Return the state instance that has given name.

        :param name: a state name
        :return: a *StateMixin* that has the same name or None
        :raise StatechartError: if state does not exist
        """
        try:
            return self._states[name]
        except KeyError as e:
            raise StatechartError('State {} does not exist'.format(name)) from e

    def parent_for(self, name: str) -> Optional[str]:
        """
        Return the name of the parent of given state name.

        :param name: a state name
        :return: its parent name, or None.
        :raise StatechartError: if state does not exist
        """
        try:
            return self._parent[name]
        except KeyError as e:
            raise StatechartError('State {} does not exist'.format(name)) from e

    def children_for(self, name: str) -> List[str]:
        """
        Return the names of the children of the given state.

        :param name: a state name
        :return: a (possibly empty) list of children
        :raise StatechartError: if state does not exist
        """
        self.state_for(name)  # Raise StatechartError if state does not exist

        return self._children[name]

    def ancestors_for(self, name: str) -> List[str]:
        """
        Return an ordered list of ancestors for the given state.
        Ancestors are ordered by decreasing depth.

        :param name: name of the state
        :return: state's ancestors
        :raise StatechartError: if state does not exist
        """
        self.state_for(name)  # Raise StatechartError if state does not exist

        ancestors = []
        parent = self._parent[name]
        while parent:
            ancestors.append(parent)
            parent = self._parent[parent]
        return ancestors

    def descendants_for(self, name: str) -> List[str]:
        """
        Return an ordered list of descendants for the given state.
        Descendants are ordered by increasing depth.

        :param name: name of the state
        :return: state's descendants
        :raise StatechartError: if state does not exist
        """
        self.state_for(name)  # Raise StatechartError if state does not exist

        descendants = []
        states_to_consider = [name]
        while states_to_consider:
            name = states_to_consider.pop(0)
            for child in self.children_for(name):
                states_to_consider.append(child)
                descendants.append(child)
        return descendants

    def depth_for(self, name: str) -> int:
        """
        Return the depth of given state (1-indexed).

        :param name: name of the state
        :return: state depth
        :raise StatechartError: if state does not exist
        """
        self.state_for(name)  # Raise StatechartError if state does not exist

        ancestors = self.ancestors_for(name)
        return len(ancestors) + 1

    def least_common_ancestor(self, name_first: str, name_second: str) -> Optional[str]:
        """
        Return the deepest common ancestor for *s1* and *s2*, or *None* if
        there is no common ancestor except root (top-level) state.

        :param name_first: name of first state
        :param name_second: name of second state
        :return: name of deepest common ancestor or *None*
        :raise StatechartError: if state does not exist
        """
        self.state_for(name_first)  # Raise StatechartError if state does not exist
        self.state_for(name_second)

        s1_anc = self.ancestors_for(name_first)
        s2_anc = self.ancestors_for(name_second)
        for state in s1_anc:
            if state in s2_anc:
                return state
        return None

    def leaf_for(self, names: Iterable[str]) -> List[str]:
        """
        Return the leaves of *names*.

        Considering the list of states names in *names*, return a list containing each
        element of *names* such that this element has no descendant in *names*.

        :param names: a list of state names
        :return: the names of the leaves in *names*
        :raise StatechartError: if a state does not exist
        """
        leaves = []  # type: List[str]
        names = set(names)  # Lookups in set are more efficient

        for name in names:
            for descendant in self.descendants_for(name):  # Raise a StatechartError if it does not exist!
                if descendant in names:
                    break
            else:  # no break occurs
                leaves.append(name)
        return leaves

    # ######### TRANSITIONS ##########

    @property
    def transitions(self):
        """
        List of available transitions
        """
        return list(self._transitions)

    def add_transition(self, transition: Transition) -> None:
        """
        Register given transition and register it on the source state

        :param transition: transition to add
        :raise StatechartError:
        """
        # Check that source state is known
        try:
            from_state = self.state_for(transition.source)
        except StatechartError as e:
            raise StatechartError('Unknown source state for {}'.format(transition)) from e

        # Check that source state is a TransactionStateMixin
        if not isinstance(from_state, TransitionStateMixin):
            raise StatechartError('Cannot add {} on {}'.format(transition, from_state))

        # Check either internal OR target state is known
        if transition.target is not None and transition.target not in self._states:
            raise StatechartError('Unknown target state for {}'.format(transition))

        self._transitions.append(transition)

    def remove_transition(self, transition: Transition) -> None:
        """
        Remove given transitions.

        :param transition: a *Transition* instance
        :raise StatechartError: if transition is not registered
        """
        try:
            self._transitions.remove(transition)
        except ValueError:
            raise StatechartError('Transition {} does not exist'.format(transition))

    def rotate_transition(self, transition: Transition, new_source: str='', new_target: Optional[str]='') -> None:
        """
        Rotate given transition.

        You MUST specify either *new_source* (a valid state name) or *new_target* (a valid state name or None) or both.

        :param transition: a *Transition* instance
        :param new_source: a state name
        :param new_target: a state name or None
        :raise StatechartError: if given transition or a given state does not exist.
        """
        # Check that either new_source or new_target is set
        if new_source == new_target == '':
            raise ValueError('You must at least specify the new source or new target')

        # Check that transition exists
        if transition not in self._transitions:
            raise StatechartError('Unknown transition {}'.format(transition))

        # Rotate using source
        if new_source != '':
            new_source_state = self.state_for(new_source)
            if not isinstance(new_source_state, TransitionStateMixin):
                raise StatechartError('{} cannot have transitions'.format(new_source_state))
            assert isinstance(new_source_state, StateMixin)
            transition._source = new_source_state.name

        # Rotate using target
        if new_target != '':
            if new_target is None:
                transition._target = None
            else:
                new_target_state = self.state_for(new_target)
                transition._target = new_target_state.name

    def transitions_from(self, source: str) -> List[Transition]:
        """
        Return the list of transitions whose source is given name.

        :param source: name of source state
        :return: a list of *Transition* instances
        :raise StatechartError: if state does not exist
        """
        self.state_for(source)  # Raise StatechartError if state does not exist

        transitions = []
        for transition in self._transitions:
            if transition.source == source:
                transitions.append(transition)
        return transitions

    def transitions_to(self, target: str) -> List[Transition]:
        """
        Return the list of transitions whose target is given name.
        Internal transitions are returned too.

        :param target: name of target state
        :return: a list of *Transition* instances
        :raise StatechartError: if state does not exist
        """
        self.state_for(target)  # Raise StatechartError if state does not exist

        transitions = []
        for transition in self._transitions:
            if transition.target == target or (transition.target is None and transition.source == target):
                transitions.append(transition)
        return transitions

    def transitions_with(self, event: str) -> List[Transition]:
        """
        Return the list of transitions that can be triggered by given event name.

        :param event: name of the event
        :return: a list of *Transition* instances
        """
        transitions = []
        for transition in self._transitions:
            if transition.event == event:
                transitions.append(transition)
        return transitions

    # ######### EVENTS ##########

    def events_for(self, name_or_names: Union[str, List[str]]=None) -> List[str]:
        """
        Return a list containing the name of every event that guards a transition in this statechart.

        If *name_or_names* is specified, it must be the name of a state (or a list of such names).
        Only transitions that have a source state from this list will be considered.
        By default, the list contains all the states.

        :param name_or_names: *None*, a state name or a list of state names.
        :return: A list of event names
        """
        if name_or_names is None:
            states = list(self._states.keys())
        elif isinstance(name_or_names, str):
            states = [name_or_names]
        else:
            states = name_or_names

        states = cast(List[str], states)
        names = set()
        for state in states:
            for transition in self.transitions_from(state):
                if transition.event:
                    names.add(transition.event)
        return sorted(names)

    # ######### STRUCTURAL CHANGES ##########

    def add_state(self, state: StateMixin, parent: Optional[str]) -> None:
        """
        Add given state (a *StateMixin* instance) on given parent (its name as an *str*).
        If given state should be use as a root state, set *parent* to None.

        :param state: state to add
        :param parent: name of its parent, or None
        :raise StatechartError:
        """
        # Check state has a name
        if state.name is None:
            raise StatechartError('State {} must have a name'.format(state))

        # Check name unicity
        if state.name in self._states.keys():
            raise StatechartError('State {} already exists!'.format(state))

        if not parent:
            # Check root state
            if self.root:
                raise StatechartError('Root already defined, {} should declare an existing parent state'.format(state))
        else:
            parent_state = self.state_for(parent)

            # Check that parent exists
            if not parent_state:
                raise StatechartError('Parent "{}" of {} does not exist!'.format(parent, state))

            # Check that parent is a CompositeStateMixin.
            if not isinstance(parent_state, CompositeStateMixin):
                raise StatechartError('{} cannot be used as a parent for {}'.format(parent_state, state))

            # If state is an HistoryState, its parent must be a CompoundState
            if isinstance(state, HistoryStateMixin) and not isinstance(parent_state, CompoundState):
                raise StatechartError('{} cannot be used as a parent for {}'.format(parent_state, state))

        # Save state
        self._states[state.name] = state
        self._parent[state.name] = parent
        self._children[state.name] = []
        self._children[parent].append(state.name)

    def remove_state(self, name: str) -> None:
        """
        Remove given state.

        The transitions that involve this state will also be removed.
        If the state is the target of an *initial* or *memory* property, their value will be set to None.
        If the state has children, they will be removed too.

        :param name: name of a state
        :raise StatechartError:
        """
        state = self.state_for(name)

        # Remove children
        for child in list(self.children_for(state.name)):
            self.remove_state(child)

        # Remove transitions
        for transition in list(self.transitions):  # Make a copy!
            if transition.source == state.name or transition.target == state.name:
                self.remove_transition(transition)

        # Remove compoundstate's initial and historystate's memory
        for o_state in self._states.values():
            if isinstance(o_state, CompoundState):
                o_state.initial = None
            elif isinstance(o_state, HistoryStateMixin):
                o_state.memory = None

        # Remove state
        self._states.pop(name)
        parent = self._parent.pop(name)
        self._children.pop(name)

        self._children[parent].remove(name)

    def rename_state(self, old_name: str, new_name: str) -> None:
        """
        Change state name, and adapt transitions, initial state, memory, etc.

        :param old_name: old name of the state
        :param new_name: new name of the state
        """
        if old_name == new_name:
            return
        if new_name in self._states:
            raise StatechartError('State {} already exists!'.format(new_name))

        # Check state exists
        state = self.state_for(old_name)

        # Change transitions
        for transition in self.transitions:
            if transition.source == old_name:
                if transition.internal:
                    transition._target = new_name
                transition._source = new_name

            if transition.target == old_name:
                transition._target = new_name

        for other_state in self._states.values():
            # Change initial (CompoundState)
            if isinstance(other_state, CompoundState):
                if other_state.initial == old_name:
                    other_state.initial = new_name

            # Change memory (HistoryState)
            if isinstance(other_state, HistoryStateMixin):
                if other_state.memory == old_name:
                    other_state.memory = new_name

            # Adapt parent
            if self._parent[other_state.name] == old_name:
                self._parent[other_state.name] = new_name

        # Adapt structures
        parent_name = self._parent[old_name]
        self._children[parent_name].remove(old_name)
        self._children[parent_name].append(new_name)

        self._states[new_name] = self._states.pop(old_name)
        self._parent[new_name] = self._parent.pop(old_name)
        self._children[new_name] = self._children.pop(old_name)

        # Rename state!
        state._name = new_name

    def move_state(self, name: str, new_parent: str) -> None:
        """
        Move given state (and its children) such that its new parent is *new_parent*.

        Notice that a state cannot be moved inside itself or inside one of its descendants.
        If the state to move is the target of an *initial* or *memory* property of its parent,
        this property will be set to None. The same occurs if given state is an history state.

        :param name: name of the state to move
        :param new_parent: name of the new parent
        """
        # Check that both states exist
        state = self.state_for(name)
        self.state_for(new_parent)

        # Check that parent is not a descendant (or self) of given state
        if new_parent in [name] + self.descendants_for(name):
            raise StatechartError('State {} cannot be moved into itself or one of its descendants.'.format(state))

        # Change its parent and register state as a child
        old_parent = self.parent_for(name)
        self._parent[name] = new_parent
        self._children[old_parent].remove(name)
        self._children.setdefault(new_parent, []).append(name)

        # Check memory property
        if isinstance(state, HistoryStateMixin):
            state.memory = None

        for other_state in self._states.values():
            # Change initial (CompoundState)
            if isinstance(other_state, CompoundState):
                if other_state.initial == name:
                    other_state.initial = None

            # Change memory (HistoryState)
            if isinstance(other_state, HistoryStateMixin):
                if other_state.memory == name:
                    other_state.memory = None

    def copy_from_statechart(self, statechart: 'Statechart', *, source: str, replace: str,
                             renaming_func: Callable[[str], str]=lambda s: s) -> None:
        """
        Copy (a part of) given *statechart* into current one.

        Copy *source* state, all its descendants and all involved transitions from *statechart* into current statechart.
        The *source* state will override *replace* state (but will be renamed to *replace*), and all its descendants in
        *statechart* will be copied into current statechart.
        All the transitions that are involved in the process must be fully contained in *source* state (ie. for all
        transition T: S->T, if S (resp. T) is a descendant-or-self of *source*, then T (resp. S) must be
        a descendant-or-self of *source*).

        If necessary, callable *renaming_func* can be provided. This function should accept a (state) name and return a
        (new state) name. Use *renaming_func* to avoid conflicting names in target statechart.

        :param statechart: Source statechart from which states will be copied.
        :param source: Name of the source state.
        :param replace: Name of the target state. Should refer to a StateMixin with no child.
        :param renaming_func: Optional callable to resolve conflicting names.
        """
        if len(self.children_for(replace)) > 0:
            raise StatechartError('State {} cannot be replaced while it has children.'.format(replace))

        statechart_copy = deepcopy(statechart)  # type: Statechart

        # Rename and copy states
        statechart_copy.rename_state(source, replace)
        source_name = replace  # For lisibility
        self._states[replace] = statechart_copy.state_for(source_name)
        for name in statechart_copy.descendants_for(source_name):
            new_name = renaming_func(name)
            # May raise a StatechartError if names collides in source statechart.
            # Possible workaround: first rename all states to uid, then rename to new names
            statechart_copy.rename_state(name, new_name)
            self.add_state(statechart_copy.state_for(new_name), statechart_copy.parent_for(new_name))

        # Copy transitions
        for name in [source_name] + statechart_copy.descendants_for(source_name):
            transitions = set(statechart_copy.transitions_from(name) + statechart_copy.transitions_to(name))
            for transition in transitions:
                try:
                    self.add_transition(transition)
                except StatechartError as e:
                    raise StatechartError('Cannot copy {} because transition {} is not contained in {}'.
                                          format(transition.source, transition, source)) from e

    # ######### VALIDATION ##########

    def _validate_compoundstate_initial(self) -> bool:
        """
        Checks that every *CompoundState*'s initial state refer to one of its children

        :return: True
        :raise StatechartError:
        """
        for name, state in self._states.items():
            if isinstance(state, CompoundState) and state.initial:
                if state.initial not in self._states:
                    raise StatechartError('Initial state {} of {} does not exist'.format(state.initial, state))
                if state.initial not in self.children_for(name):
                    raise StatechartError('Initial state {} of {} must be a child state'.format(state.initial, state))
        return True

    def _validate_historystate_memory(self) -> bool:
        """
        Checks that every *HistoryStateMixin*'s memory refer to one of its parent's children, except itself.

        :return: True
        :raise StatechartError:
        """
        for name, state in self._states.items():
            if isinstance(state, HistoryStateMixin):
                memory = state.memory
                if memory is None:
                    continue
                if memory == name:
                    raise StatechartError('Initial memory {} of {} cannot target itself'.format(state.memory, state))
                if memory not in self._states:
                    raise StatechartError('Initial memory {} of {} does not exist'.format(state.memory, state))
                if state.memory not in self.children_for(self.parent_for(name)):
                    raise StatechartError(
                        'Initial memory {} of {} must be a parent\'s child'.format(state.memory, state)
                    )
        return True

    def validate(self) -> bool:
        """
        Checks that every *CompoundState*'s initial state refer to one of its children
        Checks that every *HistoryStateMixin*'s memory refer to one of its parent's children

        :return: True
        :raise StatechartError:
        """
        self._validate_compoundstate_initial()
        self._validate_historystate_memory()

        return True



import argparse
import sys

from ..io import import_from_yaml
from .wrappers import execute_bdd


def cli(args=None) -> int:
    parser = argparse.ArgumentParser(prog='sismic-bdd',
                                     description='Command-line utility to execute Gherkin feature files using Behave.\n'
                                                 'Extra parameters will be passed to Behave.')

    parser.add_argument('statechart', metavar='statechart', type=str,
                        help='A YAML file describing a statechart')
    parser.add_argument('--features', metavar='features', nargs='+', type=str, required=True,
                        help='A list of files containing features')
    parser.add_argument('--steps', metavar='steps', nargs='+', type=str,
                        help='A list of files containing steps implementation')
    parser.add_argument('--properties', metavar='properties', nargs='+', type=str,
                        help='A list of filepaths pointing to YAML property statecharts. They will be checked at runtime following a fail fast approach.')
    parser.add_argument('--show-steps', action='store_true', default=False,
                        help='Display a list of available steps (equivalent to Behave\'s --steps parameter')
    parser.add_argument('--debug-on-error', action='store_true', default=False,
                        help='Drop in a debugger in case of step failure (ipdb if available)')

    args, parameters = parser.parse_known_args(args)
    if args.show_steps:
        parameters.append('--steps')

    statechart = import_from_yaml(filepath=args.statechart)

    property_statecharts = []
    for property_statechart in args.properties or []:
        property_statecharts.append(import_from_yaml(filepath=property_statechart))

    return execute_bdd(
        statechart,
        args.features,
        step_filepaths=args.steps,
        property_statecharts=property_statecharts,
        debug_on_error=args.debug_on_error,
        behave_parameters=parameters
    )


if __name__ == '__main__':
    sys.exit(cli())


from sismic.interpreter import Event


def test_writer(writer):
    writer.queue(
        Event('keyPress', key='bonjour '),
        Event('toggle'),
        Event('keyPress', key='a '),
        Event('toggle'),
        Event('toggle_bold'),
        Event('keyPress', key='tous !'),
        Event('leave')
    )

    writer.execute()
    assert writer.final
    assert writer.context['output'] == ['bonjour ', '[b]', '[i]', 'a ', '[/b]', '[/i]', '[b]', 'tous !', '[/b]']


class TestElevator:
    def test_init(self, elevator):
        assert elevator.configuration == []

        # Stabilisation
        elevator.execute_once()

        assert set(elevator.configuration) == {'active', 'floorListener', 'floorSelecting',
                                               'movingElevator', 'doorsOpen'}

    def test_floor_selection(self, elevator):
        # Stabilisation
        elevator.execute_once()

        elevator.queue(Event('floorSelected', floor=4)).execute_once()
        assert elevator.context['destination'] == 4

        elevator.execute_once()
        assert set(elevator.configuration) == {'active', 'active', 'doorsClosed', 'floorListener', 'floorSelecting', 'movingElevator'}

    def test_floor_selected_and_reached(self, elevator):
        # Stabilisation
        elevator.execute_once()

        elevator.queue(Event('floorSelected', floor=4)).execute()

        assert elevator.context['current'] == 4

        elevator.time += 10
        elevator.execute()

        assert 'doorsOpen' in elevator.configuration
        assert elevator.context['current'] == 0


class TestRemoteElevator:
    def test_button(self, elevator, remote_elevator):
        assert elevator.context['current'] == 0

        remote_elevator.queue(Event('button_2_pushed')).execute()

        event = elevator._external_events[-1]
        assert event.name == 'floorSelected'
        assert event.data['floor'] == 2

        elevator.execute()
        assert elevator.context['current'] == 2

    def test_button_0_on_groundfloor(self, elevator, remote_elevator):
        assert elevator.context['current'] == 0

        remote_elevator.queue(Event('button_0_pushed')).execute()
        elevator.execute()

        assert elevator.context['current'] == 0


class TestMicrowave:
    def test_lamp_on(self, microwave):
        microwave.execute_once()
        microwave.queue(Event('door_opened'))

        assert microwave.execute_once().sent_events[0].name == 'lamp_switch_on'

    def test_heating_on(self, microwave):
        microwave.execute_once()
        microwave.queue(
            Event('door_opened'),
            Event('item_placed'),
            Event('door_closed'),
            Event('timer_inc')
        ).execute()

        microwave.queue(Event('cooking_start'))
        sent_events = []
        for step in microwave.execute():
            sent_events.extend(step.sent_events)

        assert Event('heating_on') in sent_events
        assert Event('lamp_switch_on') in sent_events
        assert Event('turntable_start') in sent_events


import pytest

from sismic.exceptions import StatechartError
from sismic.model import Statechart, Transition, CompoundState, BasicState
from sismic.interpreter import Event


class TestEvents:
    def test_create_event(self):
        assert Event('hello') == Event(name='hello')

    def test_empty_event(self):
        with pytest.raises(TypeError):
            Event()

    def test_parametrized_events(self):
        event = Event('test', a=1, b=2, c=3)

        assert event.data == {'a': 1, 'b': 2, 'c': 3}
        assert event.a == 1
        assert event.b == 2
        assert event.c == 3
        assert event.name == 'test'

        with pytest.raises(AttributeError):
            _ = event.d

        with pytest.raises(KeyError):
            _ = event.data['d']

    def test_cannot_use_name_as_parameter(self):
        with pytest.raises(TypeError):
            Event('test', name='fail')


class TestStatechartTraveral:
    def test_parent(self, composite_statechart):
        assert composite_statechart.parent_for('s2') == 'root'
        assert composite_statechart.parent_for('root') is None

        with pytest.raises(StatechartError):
            composite_statechart.parent_for('unknown')

    def test_children(self, composite_statechart):
        assert set(composite_statechart.children_for('root')) == {'s1', 's2'}
        assert composite_statechart.children_for('s1b2') == []

    def test_ancestors(self, composite_statechart):
        assert composite_statechart.ancestors_for('s2') == ['root']
        assert composite_statechart.ancestors_for('s1a') == ['s1', 'root']
        assert composite_statechart.ancestors_for('s1b1') == ['s1b', 's1', 'root']

    def test_descendants(self, composite_statechart):
        assert composite_statechart.descendants_for('s2') == []
        assert set(composite_statechart.descendants_for('s1')) == {'s1a', 's1b', 's1b1', 's1b2'}

    def test_depth(self, composite_statechart):
        assert composite_statechart.depth_for('root') == 1
        assert composite_statechart.depth_for('s1') == 2
        assert composite_statechart.depth_for('s1a') == 3
        assert composite_statechart.depth_for('s1b1') == 4
        assert composite_statechart.depth_for('s2') == 2
        assert composite_statechart.depth_for('s1b2') == 4

    def test_lca(self, composite_statechart):
        assert composite_statechart.least_common_ancestor('s1', 's2') == 'root'
        assert composite_statechart.least_common_ancestor('s1', 's1a') == 'root'
        assert composite_statechart.least_common_ancestor('s1a', 's1b') == 's1'
        assert composite_statechart.least_common_ancestor('s1a', 's1b1') == 's1'

    def test_leaf(self, composite_statechart):
        assert sorted(composite_statechart.leaf_for([])) == []
        assert sorted(composite_statechart.leaf_for(['s1'])) == ['s1']
        assert sorted(composite_statechart.leaf_for(['s1', 's2'])) == ['s1', 's2']
        assert sorted(composite_statechart.leaf_for(['s1', 's1b1', 's2'])) == ['s1b1', 's2']
        assert sorted(composite_statechart.leaf_for(['s1', 's1b', 's1b1'])) == ['s1b1']

    def test_events_for(self, composite_statechart):
        assert set(composite_statechart.events_for()) == {'click', 'close', 'validate'}
        assert set(composite_statechart.events_for('s1b1')) == {'validate'}
        assert set(composite_statechart.events_for(['s1b1', 's1b'])) == {'click', 'validate'}

    def test_name_collision(self):
        root = CompoundState('root', 'a')
        sc = Statechart('test')
        sc.add_state(root, None)
        s1 = BasicState('a')
        s2 = BasicState('a')
        sc.add_state(s1, parent='root')

        with pytest.raises(StatechartError) as e:
            sc.add_state(s2, parent='root')
        assert 'already exists!' in str(e.value)

    def test_name_is_none(self):
        sc = Statechart('test')
        state = BasicState(name=None)
        with pytest.raises(StatechartError) as e:
            sc.add_state(state, None)
        assert 'must have a name' in str(e.value)

    def test_root_already_defined(self):
        root = CompoundState('root', 'a')
        sc = Statechart('test')
        sc.add_state(root, None)

        with pytest.raises(StatechartError) as e:
            sc.add_state(root, None)
        assert 'already exists!' in str(e.value)


class TestStatechartValidate:
    # REMARK: "Positive" tests are already done during io.import_from_yaml!
    def test_history_memory_is_not_a_child(self, history_statechart):
        history_statechart.state_for('loop.H').memory = 'pause'
        with pytest.raises(StatechartError) as e:
            history_statechart._validate_historystate_memory()
        assert 'Initial memory' in str(e.value)
        assert 'must be a parent\'s child' in str(e.value)

    def test_history_memory_unknown(self, history_statechart):
        history_statechart.state_for('loop.H').memory = 'unknown'
        with pytest.raises(StatechartError) as e:
            history_statechart._validate_historystate_memory()
        assert 'Initial memory' in str(e.value)
        assert 'does not exist' in str(e.value)

    def test_history_memory_self(self, history_statechart):
        history_statechart.state_for('loop.H').memory = 'loop.H'
        with pytest.raises(StatechartError) as e:
            history_statechart._validate_historystate_memory()
        assert 'Initial memory' in str(e.value)
        assert 'cannot target itself' in str(e.value)

    def test_compound_initial_not_a_child(self, composite_statechart):
        composite_statechart.state_for('s1b').initial = 's1'
        with pytest.raises(StatechartError) as e:
            composite_statechart._validate_compoundstate_initial()
        assert 'Initial state' in str(e.value)
        assert 'must be a child state' in str(e.value)

    def test_compound_initial_unknown(self, composite_statechart):
        composite_statechart.state_for('s1b').initial = 'unknown'
        with pytest.raises(StatechartError) as e:
            composite_statechart._validate_compoundstate_initial()
        assert 'Initial state' in str(e.value)
        assert 'does not exist' in str(e.value)


class TestTransition:
    def test_transitions_from(self, internal_statechart):
        assert internal_statechart.transitions_from('root') == []
        assert len(internal_statechart.transitions_from('active')) == 2
        assert len(internal_statechart.transitions_from('s1')) == 1

        for transition in internal_statechart.transitions_from('active'):
            assert transition.source == 'active'

        with pytest.raises(StatechartError) as e:
            internal_statechart.transitions_from('unknown')
        assert 'does not exist' in str(e.value)

    def test_transitions_to(self, internal_statechart):
        assert internal_statechart.transitions_to('root') == []
        assert len(internal_statechart.transitions_to('s1')) == 1
        assert len(internal_statechart.transitions_to('s2')) == 2

        for transition in internal_statechart.transitions_from('s2'):
            assert transition.target == 's2'

        with pytest.raises(StatechartError) as e:
            internal_statechart.transitions_to('unknown')
        assert 'does not exist' in str(e.value)

    def test_transitions_with(self, internal_statechart):
        assert len(internal_statechart.transitions_with('next')) == 1
        assert len(internal_statechart.transitions_with('unknown')) == 0

    def test_add_transition(self, internal_statechart):
        with pytest.raises(StatechartError) as e:
            internal_statechart.add_transition(Transition('s2'))
        assert 'Cannot add' in str(e.value)

        internal_statechart.add_transition(Transition('s1'))
        assert len(internal_statechart.transitions_to('s1')) == 2

    def test_rotate_source(self, internal_statechart):
        tr = next(t for t in internal_statechart.transitions if t.source == 's1')
        internal_statechart.rotate_transition(tr, new_source='s1')
        assert tr.source == 's1'

        internal_statechart.rotate_transition(tr, new_source='active')
        assert tr.source == 'active'

        with pytest.raises(StatechartError) as e:
            internal_statechart.rotate_transition(tr, new_source=None)
        assert 'State' in str(e.value)
        assert 'does not exist' in str(e.value)

        with pytest.raises(StatechartError) as e:
            internal_statechart.rotate_transition(tr, new_source='s2')
        assert 'cannot have transitions' in str(e.value)

        with pytest.raises(StatechartError) as e:
            internal_statechart.rotate_transition(tr, new_source='unknown')
        assert 'State' in str(e.value)
        assert 'does not exist' in str(e.value)

        internal_statechart.validate()

    def test_rotate_target(self, internal_statechart):
        tr = next(t for t in internal_statechart.transitions if t.source == 's1')
        internal_statechart.rotate_transition(tr, new_target='s2')
        assert tr.target == 's2'

        internal_statechart.rotate_transition(tr, new_target='active')
        assert tr.target == 'active'

        internal_statechart.rotate_transition(tr, new_target=None)
        assert tr.target is None
        assert tr.internal

        with pytest.raises(StatechartError) as e:
            internal_statechart.rotate_transition(tr, new_target='unknown')
        assert 'State' in str(e.value)
        assert 'does not exist' in str(e.value)

        internal_statechart.validate()

    def test_rotate_both(self, internal_statechart):
        tr = next(t for t in internal_statechart.transitions if t.source == 's1')

        internal_statechart.rotate_transition(tr, new_source='s1', new_target='s2')
        assert tr.source == 's1'
        assert tr.target == 's2'
        internal_statechart.validate()

    def test_rotate_both_unexisting(self, internal_statechart):
        tr = next(t for t in internal_statechart.transitions if t.source == 's1')

        with pytest.raises(ValueError):
            internal_statechart.rotate_transition(tr)

        with pytest.raises(StatechartError) as e:
            internal_statechart.rotate_transition(tr, new_source=None, new_target=None)
        assert 'State' in str(e.value)
        assert 'does not exist' in str(e.value)

        with pytest.raises(StatechartError) as e:
            internal_statechart.rotate_transition(tr, new_source='s2', new_target='s2')
        assert 'State' in str(e.value)
        assert 'cannot have transitions' in str(e.value)

        internal_statechart.validate()

    def test_rotate_both_with_internal(self, internal_statechart):
        tr = next(t for t in internal_statechart.transitions if t.source == 's1')

        internal_statechart.rotate_transition(tr, new_source='s1', new_target=None)
        assert tr.source == 's1'
        assert tr.target is None
        assert tr.internal
        internal_statechart.validate()

    def test_remove_existing_transition(self, internal_statechart):
        transitions = internal_statechart.transitions
        for transition in transitions:
            internal_statechart.remove_transition(transition)
        assert internal_statechart.transitions == []
        internal_statechart.validate()

    def test_remove_unexisting_transition(self, internal_statechart):
        with pytest.raises(StatechartError) as e:
            internal_statechart.remove_transition(None)
        assert 'Transition' in str(e.value)
        assert 'does not exist' in str(e.value)

        with pytest.raises(StatechartError) as e:
            internal_statechart.remove_transition(Transition('a', 'b'))
        assert 'Transition' in str(e.value)
        assert 'does not exist' in str(e.value)

        internal_statechart.validate()


class TestStatechartStates:
    def test_remove_existing_state(self, internal_statechart):
        internal_statechart.remove_state('active')

        assert 'active' not in internal_statechart.states
        nb_transitions = 0
        for transition in internal_statechart.transitions:
            if transition.source == 'active':
                nb_transitions += 1
        assert nb_transitions == 0

        nb_transitions = 0
        for transition in internal_statechart.transitions:
            if transition.target == 'active':
                nb_transitions += 1
        assert nb_transitions == 0
        internal_statechart.validate()

    def test_remove_unregister_parent_children(self, internal_statechart):
        internal_statechart.remove_state('s1')
        assert 's1' not in internal_statechart.children_for('active')
        internal_statechart.validate()

    def test_remove_unexisting_state(self, internal_statechart):
        with pytest.raises(StatechartError) as e:
            internal_statechart.remove_state('unknown')
        assert 'State' in str(e.value)
        assert 'does not exist' in str(e.value)

        internal_statechart.validate()

    def test_remove_root_state(self, internal_statechart):
        internal_statechart.remove_state('root')
        assert internal_statechart.transitions == []
        assert internal_statechart.states == []
        assert internal_statechart.root is None
        internal_statechart.validate()

    def test_remove_nested_states(self, composite_statechart):
        composite_statechart.remove_state('s1')
        assert 's1a' not in composite_statechart.states
        composite_statechart.validate()

    def test_remove_nested_states_continued(self, composite_statechart):
        composite_statechart.remove_state('s1a')
        assert 's1a' not in composite_statechart.states
        composite_statechart.validate()

    def test_rename_unexisting_state(self, internal_statechart):
        with pytest.raises(StatechartError) as e:
            internal_statechart.rename_state('unknown', 's3')
        assert 'State' in str(e.value)
        assert 'does not exist' in str(e.value)

        internal_statechart.validate()

    def test_do_not_change_name(self, internal_statechart):
        internal_statechart.rename_state('s2', 's2')

        assert 's2' in internal_statechart.states
        assert internal_statechart.parent_for('s2') == 'root'
        assert 's2' in internal_statechart.children_for('root')

        internal_statechart.validate()

    def test_rename_to_an_existing_state(self, internal_statechart):
        with pytest.raises(StatechartError) as e:
            internal_statechart.rename_state('s2', 's1')
        assert 'State' in str(e.value)
        assert 'already exists!' in str(e.value)

        internal_statechart.validate()

    def test_rename_old_disappears(self, composite_statechart):
        composite_statechart.rename_state('s1', 'new s1')

        assert 's1' not in composite_statechart.states
        assert 's1' != composite_statechart.parent_for('s1a')
        assert 's1' not in composite_statechart.children_for('root')

        with pytest.raises(StatechartError) as e:
            composite_statechart.state_for('s1')
        assert 'State' in str(e.value)
        assert 'does not exist' in str(e.value)

        with pytest.raises(StatechartError) as e:
            composite_statechart.children_for('s1')
        assert 'State' in str(e.value)
        assert 'does not exist' in str(e.value)

        with pytest.raises(StatechartError) as e:
            composite_statechart.parent_for('s1')
        assert 'State' in str(e.value)
        assert 'does not exist' in str(e.value)

        composite_statechart.validate()

    def test_rename_change_state_name(self, composite_statechart):
        state = composite_statechart.state_for('s1')
        composite_statechart.rename_state('s1', 'new s1')
        assert state.name == 'new s1'

    def test_rename_new_appears(self, composite_statechart):
        composite_statechart.rename_state('s1', 'new s1')

        assert 'new s1' in composite_statechart.states
        assert 'new s1' == composite_statechart.parent_for('s1a')
        assert 'new s1' in composite_statechart.children_for('root')

        composite_statechart.state_for('new s1')
        composite_statechart.children_for('new s1')
        composite_statechart.parent_for('new s1')
        composite_statechart.validate()

    def test_rename_adapt_transitions(self, internal_statechart):
        internal_statechart.rename_state('s1', 's3')

        nb_transitions = 0
        for transition in internal_statechart.transitions:
            if transition.source == 's1':
                nb_transitions += 1
        assert nb_transitions == 0

        nb_transitions = 0
        for transition in internal_statechart.transitions:
            if transition.target == 's1':
                nb_transitions += 1
        assert nb_transitions == 0

        assert len(internal_statechart.transitions_from('s3')) == 1
        assert len(internal_statechart.transitions_to('s2')) == 2
        internal_statechart.validate()

    def test_rename_root(self, internal_statechart):
        internal_statechart.rename_state('root', 'new root')

        assert 'new root' in internal_statechart.states
        assert 'root' not in internal_statechart.states

        assert internal_statechart.parent_for('new root') is None
        assert internal_statechart.parent_for('active') == 'new root'
        assert 'active' in internal_statechart.children_for('new root')

        internal_statechart.validate()

    def test_rename_change_initial(self, internal_statechart):
        internal_statechart.rename_state('active', 's3')
        assert internal_statechart.state_for('root').initial == 's3'
        internal_statechart.validate()

    def test_rename_change_memory(self, history_statechart):
        history_statechart.state_for('loop.H').memory = 's1'
        history_statechart.rename_state('s1', 's4')
        assert history_statechart.state_for('loop.H').memory == 's4'
        history_statechart.validate()

    def test_move_simple(self, composite_statechart):
        composite_statechart.move_state('s1b2', 's1b1')
        assert 's1b2' in composite_statechart.children_for('s1b1')
        assert 's1b2' not in composite_statechart.children_for('s1b')
        assert 's1b1' == composite_statechart.parent_for('s1b2')
        composite_statechart.validate()

    def test_move_composite(self, composite_statechart):
        composite_statechart.move_state('s1b', 's1a')
        assert 's1b' in composite_statechart.children_for('s1a')
        assert 's1b' not in composite_statechart.children_for('s1')
        assert 's1a' == composite_statechart.parent_for('s1b')
        composite_statechart.validate()

    def test_move_to_descendant(self, composite_statechart):
        with pytest.raises(StatechartError) as e:
            composite_statechart.move_state('s1b', 's1b')
        assert 'State' in str(e.value)
        assert 'cannot be moved into itself or one of its descendants' in str(e.value)

        with pytest.raises(StatechartError) as e:
            composite_statechart.move_state('s1b', 's1b1')
        assert 'State' in str(e.value)
        assert 'cannot be moved into itself or one of its descendants' in str(e.value)

        with pytest.raises(StatechartError) as e:
            composite_statechart.move_state('s1', 's1b1')
        assert 'State' in str(e.value)
        assert 'cannot be moved into itself or one of its descendants' in str(e.value)

        composite_statechart.validate()

    def test_move_unexisting(self, composite_statechart):
        with pytest.raises(StatechartError) as e:
            composite_statechart.move_state('unknown', 's1b1')
        assert 'State' in str(e.value)
        assert 'does not exist' in str(e.value)

        with pytest.raises(StatechartError) as e:
            composite_statechart.move_state('s1b1', 'unknown')
        assert 'State' in str(e.value)
        assert 'does not exist' in str(e.value)

        composite_statechart.validate()

    def test_move_with_initial(self, composite_statechart):
        composite_statechart.move_state('s1a', 'root')
        assert composite_statechart.state_for('s1').initial is None
        composite_statechart.validate()

    def test_move_with_memory(self, history_statechart):
        history_statechart.state_for('loop.H').memory = 's1'
        history_statechart.move_state('s1', 's2')
        assert history_statechart.state_for('loop.H').memory is None
        history_statechart.validate()


class TestCopyFromStatechart:
    @pytest.fixture
    def modified_simple_statechart(self, simple_statechart):
        for state in simple_statechart.states:
            simple_statechart.rename_state(state, 'sc1_' + state)
        return simple_statechart

    def test_keep_other_states(self, modified_simple_statechart, composite_statechart):
        sc1_states = modified_simple_statechart.states

        composite_statechart.copy_from_statechart(modified_simple_statechart, source='sc1_root', replace='s1a')
        assert 's1a' in composite_statechart.states
        assert 'sc1_root' not in composite_statechart.states

        sc1_states.remove('sc1_root')
        for state in sc1_states:
            assert state in composite_statechart.states

    def test_keep_other_transitions(self, modified_simple_statechart, composite_statechart):
        sc1_transitions = modified_simple_statechart.transitions

        composite_statechart.copy_from_statechart(modified_simple_statechart, source='sc1_root', replace='s1b1')

        for transition in sc1_transitions:
            assert transition in composite_statechart.transitions

    def test_keep_existing_states(self, modified_simple_statechart, composite_statechart):
        sc2_states = composite_statechart.states

        composite_statechart.copy_from_statechart(modified_simple_statechart, source='sc1_root', replace='s1b1')

        for state in sc2_states:
            assert state in composite_statechart.states

    def test_keep_existing_transitions(self, modified_simple_statechart, composite_statechart):
        sc2_transitions = composite_statechart.transitions

        composite_statechart.copy_from_statechart(modified_simple_statechart, source='sc1_root', replace='s1b1')

        for transition in sc2_transitions:
            assert transition in composite_statechart.transitions

    def test_orphaned_transitions(self, modified_simple_statechart, composite_statechart):
        with pytest.raises(StatechartError) as e:
            composite_statechart.copy_from_statechart(modified_simple_statechart, source='sc1_s1', replace='s1b1')
        assert 'not contained in' in str(e.value)

        with pytest.raises(StatechartError) as e:
            composite_statechart.copy_from_statechart(modified_simple_statechart, source='sc1_final', replace='s1b1')
            assert 'not contained in' in str(e.value)

    def test_invalid_plug(self, modified_simple_statechart, composite_statechart):
        # On compound
        with pytest.raises(StatechartError) as e:
            composite_statechart.copy_from_statechart(modified_simple_statechart, source='sc1_root', replace='s1')
        assert 'children' in str(e.value)

        with pytest.raises(StatechartError) as e:
            composite_statechart.copy_from_statechart(modified_simple_statechart, source='sc1_root', replace='s1b')
            assert 'children' in str(e.value)

    def test_with_namespace(self, modified_simple_statechart, composite_statechart):
        sc1_states = modified_simple_statechart.states
        sc2_states = composite_statechart.states

        namespace = lambda s: '__' + s
        composite_statechart.copy_from_statechart(modified_simple_statechart, source='sc1_root', replace='s1a', renaming_func=namespace)

        expected_states = sc2_states + [namespace(s) for s in sc1_states]
        expected_states.remove(namespace(modified_simple_statechart.root))

        assert set(expected_states) == set(composite_statechart.states)

    def test_conflicting_names(self, modified_simple_statechart, composite_statechart):
        modified_simple_statechart.rename_state('sc1_s1', 's1')
        with pytest.raises(StatechartError) as e:
            composite_statechart.copy_from_statechart(modified_simple_statechart, source='sc1_root', replace='s1a')
        assert 'already exists' in str(e.value)



from collections import deque
from itertools import combinations, groupby
from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, Set, Union, cast, Tuple

try:
    # Because Deque was not available in early versions of typing...
    from typing import Deque
except ImportError:
    pass

from ..model import (
    MacroStep, MicroStep, Event, InternalEvent, MetaEvent,
    Statechart, Transition,
    StateMixin, FinalState, OrthogonalState, CompoundState, DeepHistoryState, ShallowHistoryState,
)

from ..code import Evaluator, PythonEvaluator
from ..exceptions import (ConflictingTransitionsError, InvariantError, PropertyStatechartError,
                          ExecutionError, NonDeterminismError, PostconditionError, PreconditionError)

__all__ = ['Interpreter']


class Interpreter:
    """
    A discrete interpreter that executes a statechart according to a semantic close to SCXML.

    :param statechart: statechart to interpret
    :param evaluator_klass: An optional callable (eg. a class) that takes an interpreter and an optional initial
        context as input and return an *Evaluator* instance that will be used to initialize the interpreter.
        By default, the *PythonEvaluator* class will be used.
    :param initial_context: an optional initial context that will be provided to the evaluator.
        By default, an empty context is provided
    :param ignore_contract: set to True to ignore contract checking during the execution.
    """

    def __init__(self, statechart: Statechart, *,
                 evaluator_klass: Callable[..., Evaluator]=PythonEvaluator,
                 initial_context: Mapping[str, Any]=None,
                 ignore_contract: bool=False) -> None:
        # Internal variables
        self._ignore_contract = ignore_contract
        self._statechart = statechart

        self._initialized = False

        # Internal clock
        self._time = 0  # type: float

        # History states memory
        self._memory = {}  # type: Dict[str, Optional[List[str]]]

        # Set of active states
        self._configuration = set()  # type: Set[str]

        # External events queue
        self._external_events = deque()  # type: Deque[Event]

        # Internal events queue
        self._internal_events = deque()  # type: Deque[InternalEvent]

        # Bound callables
        self._bound = []  # type: List[Callable[[Event], Any]]

        # Bound property statecharts
        self._bound_properties = []  # type: List[Interpreter]

        # Evaluator
        self._evaluator = evaluator_klass(self, initial_context=initial_context)  # type: ignore
        self._evaluator.execute_statechart(statechart)

    @property
    def time(self) -> float:
        """
        Time value (in seconds) for the internal clock
        """
        return self._time

    @time.setter
    def time(self, value: float):
        """
        Set the time of the internal clock

        :param value: time value (in seconds)
        """
        if self._time > value:
            raise ExecutionError('Time must be monotonic, cannot set time to {} from {}'.format(value, self._time))
        self._time = value

        # Update bound properties
        for property_statechart in self._bound_properties:
            property_statechart.time = self._time

    @property
    def configuration(self) -> List[str]:
        """
        List of active states names, ordered by depth. Ties are broken according to the lexicographic order
        on the state name.
        """
        return sorted(self._configuration, key=lambda s: (self._statechart.depth_for(s), s))

    @property
    def context(self) -> Mapping[str, Any]:
        """
        The context of execution.
        """
        return self._evaluator.context

    @property
    def final(self) -> bool:
        """
        Boolean indicating whether this interpreter is in a final configuration.
        """
        return self._initialized and len(self._configuration) == 0

    @property
    def statechart(self) -> Statechart:
        """
        Embedded statechart
        """
        return self._statechart

    def bind(self, interpreter_or_callable: Union['Interpreter', Callable[[Event], Any]]) -> None:
        """
        Bind an interpreter or a callable to the current interpreter.
        Each time an internal event is sent by this interpreter, any bound object will be called
        with the same event. If *interpreter_or_callable* is an *Interpreter* instance,  its *queue* method is called.
        This is, if *i1* and *i2* are interpreters, *i1.bind(i2)* is equivalent to *i1.bind(i2.queue)*.

        :param interpreter_or_callable: interpreter or callable to bind
        :return: *self* so it can be chained
        """
        if isinstance(interpreter_or_callable, Interpreter):
            self._bound.append(interpreter_or_callable.queue)
        else:
            self._bound.append(interpreter_or_callable)

    def bind_property_statechart(self, statechart_or_interpreter: Union[Statechart, 'Interpreter']) -> None:
        """
        Bind a property statechart to the current interpreter.
        A property statechart receives meta-events from the current interpreter depending on what happens:

         - *step started*: when a macro step starts.
         - *step ended*: when a macro step ends.
         - *event consumed*: when an event is consumed. The consumed event is exposed through the ``event`` attribute.
         - *event sent*: when an event is sent. The sent event is exposed through the ``event`` attribute.
         - *state exited*: when a state is exited. The exited state is exposed through the ``state`` attribute.
         - *state entered*: when a state is entered. The entered state is exposed through the ``state`` attribute.
         - *transition processed*: when a transition is processed. The source state, target state and the event are
           exposed respectively through the ``source``, ``target`` and ``event`` attribute.

        Additionally, MetaEvent instances that are sent from within the statechart are directly passed to all
        bound property statecharts. This allows more advanced communication and synchronisation patterns with
        the bound property statecharts.

        The internal clock of all property statecharts will be synced with the one of the current interpreter.
        As soon as a property statechart reaches a final state, a ``PropertyStatechartError`` will be raised,
        implicitly meaning that the property expressed by the corresponding property statechart is not satisfied.

        :param statechart_or_interpreter: A property statechart or an interpreter of a property statechart.
        """
        # Create interpreter if required
        if isinstance(statechart_or_interpreter, Statechart):
            interpreter = Interpreter(statechart_or_interpreter)
        else:
            interpreter = statechart_or_interpreter

        # Sync clock
        interpreter.time = self.time

        # Add to the list of properties
        self._bound_properties.append(interpreter)

    def queue(self, event_or_name: Union[str, Event], *events_or_names: Union[str, Event]) -> 'Interpreter':
        """
        Queue one or more events to the interpreter external queue.

        :param event_or_name: an *Event* instance, or the name of an event.
        :param events_or_names: additional *Event* instances, or names of events.
        :return: *self* so it can be chained.
        """
        if len(events_or_names) == 0:
            events_or_names = tuple()

        for item in [event_or_name] + list(events_or_names):
            if isinstance(item, Event):
                self._external_events.append(item)
            elif isinstance(item, str):
                self._external_events.append(Event(item))
            else:
                raise ValueError('{} is not a string nor an Event instance.'.format(item))

        return self

    def execute(self, max_steps: int = -1) -> List[MacroStep]:
        """
        Repeatedly calls *execute_once* and return a list containing
        the returned values of *execute_once*.

        Notice that this does NOT return an iterator but computes the whole list first
        before returning it.

        :param max_steps: An upper bound on the number steps that are computed and returned.
            Default is -1, no limit. Set to a positive integer to avoid infinite loops
            in the statechart execution.
        :return: A list of *MacroStep* instances
        """
        returned_steps = []
        i = 0
        macro_step = self.execute_once()
        while macro_step:
            returned_steps.append(macro_step)
            i += 1
            if 0 < max_steps == i:
                break
            macro_step = self.execute_once()
        return returned_steps

    def execute_once(self) -> Optional[MacroStep]:
        """
        Select transitions that can be fired based on available queued events, process them and stabilize
        the interpreter. When multiple transitions are selected, they are atomically processed:
        states are exited, transition is processed, states are entered, statechart is stabilized and only
        after that, the next transition is processed.

        :return: a macro step or *None* if nothing happened
        """
        # Compute steps
        computed_steps = self._compute_steps()
        
        if computed_steps is None:
            # No step (no transition, no event). However, check properties
            self._check_properties(None)
            return None

        # Notify properties
        self._notify_properties('step started')

        # Consume event if it triggered a transition
        if computed_steps[0].event is not None:
            event = self._select_event(consume=True)
            self._notify_properties('event consumed', event=event)
        else:
            event = None

        # Execute the steps
        self._evaluator.on_step_starts(event)
        executed_steps = []
        for step in computed_steps:
            executed_steps.append(self._apply_step(step))
            executed_steps.extend(self._stabilize())

        macro_step = MacroStep(time=self.time, steps=executed_steps)

        # Check state invariants
        configuration = self.configuration  # Use self.configuration to benefit from the sorting by depth
        for name in configuration:
            state = self._statechart.state_for(name)
            self._evaluate_contract_conditions(state, 'invariants', macro_step)

        # End step and check for property statechart violations
        self._notify_properties('step ended')
        self._check_properties(macro_step)

        return macro_step

    def _raise_event(self, event: Event) -> None:
        """
        Raise an event from the statechart.

        InternalEvent instances are propagated to bound interpreters as non-internal events, and added to the internal
        event queue of the current interpreter as an InternalEvent instance.

        If a MetaEvent instance is provided, it is used to notify the bound property statecharts and is neither
        propagated to bound interpreters nor added to the internal event queue of the current interpreter.

        :param event: event to be sent by the statechart.
        """
        if isinstance(event, InternalEvent):
            # Add to current interpreter's internal queue
            self._internal_events.append(event)

            # Notify bound property statecharts
            self._notify_properties('event sent', event=event)

            # Propagate event to bound callable as an "external" event
            external_event = Event(event.name, **event.data)
            for bound_callable in self._bound:
                bound_callable(external_event)
        elif isinstance(event, MetaEvent):
            # Notify bound property statecharts
            self._notify_properties(event.name, **event.data)
        else:
            raise ValueError('Only InternalEvent and MetaEvent can be sent by a statechart, not {}'.format(type(event)))

    def _notify_properties(self, event_name: str, **kwargs):
        """
        Notify the property statecharts bound to this interpreter.

        :param event_name: name of the event
        :param kwargs: additional parameters that should be made available as event parameters
        """

        # Create meta-event
        event = MetaEvent(event_name, **kwargs)

        # Send meta-event to the bound properties
        for property_statechart in self._bound_properties:
            property_statechart.queue(event)

        # Execute property statecharts
        for property_statechart in self._bound_properties:
            property_statechart.execute()

    def _check_properties(self, macro_step: Optional[MacroStep]):
        """
        Check property statecharts for failure (ie. final state is reached).

        :param macro_step: current macro step being processed
        """
        for property_statechart in self._bound_properties:
            # Check for failure
            if property_statechart.final:
                raise PropertyStatechartError(property_statechart, self.configuration, macro_step, self.context)

    def _select_event(self, *, consume: bool) -> Optional[Event]:
        """
        Return the next available event if any.
        This method prioritizes internal events over external ones.

        :param consume: Indicates whether event should be consumed.
        :return: An instance of Event or None if no event is available
        """
        # Internal events are processed first
        if len(self._internal_events) > 0:
            if consume:
                return self._internal_events.popleft()
            else:
                return self._internal_events[0]
        elif len(self._external_events) > 0:
            if consume:
                return self._external_events.popleft()
            else:
                return self._external_events[0]
        else:
            return None

    def _select_transitions(self, event: Event, states: List[str], *,
            eventless_first=True, inner_first=True) -> List[Transition]:
        """
        Select and return the transitions that are triggered, based on given event
        (or None if no event can be consumed) and given list of states. 
        
        By default, this function prioritizes eventless transitions and follows
        inner-first/source state semantics. 
        
        :param event: event to consider, possibly None.
        :param states: state names to consider. 
        :param eventless_first: True to prioritize eventless transitions.
        :param inner_first: True to follow inner-first/source state semantics.
        :return: list of triggered transitions.
        """        
        selected_transitions = []  # type: List[Transition]
        
        # List of (event_order, depth_order, transition)
        considered_transitions = []  # type: List[Tuple[int, int, Transition]]

        # Use a cache for state depth (could avoid some costly computations)
        _state_depth_cache = dict()  # type: Dict[str, int]

        # Select triggerable (based on event) transitions for considered states
        for transition in self._statechart.transitions:
            if transition.source in states:
                if transition.event is None or transition.event == getattr(event, 'name', None):
                    # Compute order based on event
                    event_order = int(transition.event is None)  # event = 0, no event = 1
                    
                    # Compute order based on depth
                    if transition.source not in _state_depth_cache:
                        _state_depth_cache[transition.source] = self._statechart.depth_for(transition.source)                    
                    depth_order = _state_depth_cache[transition.source]  # less nested first
                    
                    # Change order according to the semantics
                    event_order = -event_order if eventless_first else event_order
                    depth_order = -depth_order if inner_first else depth_order

                    considered_transitions.append((event_order, depth_order, transition))

        # Order transitions based on event and depth orderings
        considered_transitions.sort(key=lambda t: (t[0], t[1]))
        
        # Which states should be selected to satisfy depth ordering?
        ignored_state_selector = self._statechart.ancestors_for if inner_first else self._statechart.descendants_for
        ignored_states = set()  # type: Set[str]
        
        # Select transitions according to the semantics
        for eventless, transitions in groupby(considered_transitions, lambda t: t[0]):  # event order
            # Event shouldn't be exposed to guards if we're processing eventless transition
            exposed_event = None if eventless else event
            
            # If there are selected transitions (from previous group), ignore new ones
            if len(selected_transitions) > 0:
                break

            # Remember that transitions are sorted based on event/eventless and the source state depth
            for *_, transition in transitions: 
                if transition.source not in ignored_states:
                    if transition.guard is None or self._evaluator.evaluate_guard(transition, exposed_event):
                        # Add descendants/ancestors to ignored states
                        for state in ignored_state_selector(transition.source):
                            ignored_states.add(state)
                        # Add transition to the list of selected ones
                        selected_transitions.append(transition)
                        
        return selected_transitions

    def _sort_transitions(self, transitions: List[Transition]) -> List[Transition]:
        """
        Given a list of triggered transitions, return a list of transitions in an order that represents
        the order in which they have to be processed.

        :param transitions: a list of *Transition* instances
        :return: an ordered list of *Transition* instances
        :raise ExecutionError: In case of non-determinism (*NonDeterminismError*) or conflicting
            transitions (*ConflictingTransitionsError*).
        """
        if len(transitions) > 1:
            # If more than one transition, we check (1) they are from separate regions and (2) they do not conflict
            # Two transitions conflict if one of them leaves the parallel state
            for t1, t2 in combinations(transitions, 2):
                # Check (1)
                lca = self._statechart.least_common_ancestor(t1.source, t2.source)  # type: str
                lca_state = self._statechart.state_for(lca)

                # Their LCA must be an orthogonal state!
                if not isinstance(lca_state, OrthogonalState):
                    raise NonDeterminismError(
                        'Non-determinist choice between transitions {t1} and {t2}'
                        '\nConfiguration is {c}\nEvent is {e}\nTransitions are:{t}\n'
                        .format(c=self.configuration, e=t1.event, t=transitions, t1=t1, t2=t2)
                    )

                # Check (2)
                # This check must be done wrt. to LCA, as the combination of from_states could
                # come from nested parallel regions!
                for transition in [t1, t2]:
                    last_before_lca = transition.source
                    for state in self._statechart.ancestors_for(transition.source):
                        if state == lca:
                            break
                        last_before_lca = state
                    # Target must be a descendant (or self) of this state
                    if (transition.target and
                            (transition.target not in
                             [last_before_lca] + self._statechart.descendants_for(last_before_lca))):
                        raise ConflictingTransitionsError(
                            'Conflicting transitions: {t1} and {t2}'
                            '\nConfiguration is {c}\nEvent is {e}\nTransitions are:{t}\n'
                            .format(c=self.configuration, e=t1.event, t=transitions, t1=t1, t2=t2)
                        )

            # Define an arbitrary order based on the depth and the name of source states.
            transitions = sorted(transitions, key=lambda t: (-self._statechart.depth_for(t.source), t.source))

        return transitions

    def _compute_steps(self) -> Optional[List[MicroStep]]:
        """
        Compute and returns the next steps based on current configuration
        and event queues. 

        :return A (possibly None) list of steps.
        """
        # Initialization
        if not self._initialized:
            self._initialized = True
            return [MicroStep(entered_states=[self._statechart.root])]
        
        # Select transitions
        event = self._select_event(consume=False)
        transitions = self._select_transitions(event, states=self._configuration)

        # No transition can be triggered?
        if len(transitions) == 0:
            if event is None:
                # No event, no step!
                return None
            else:
                # Empty step, so that event is eventually consumed
                return [MicroStep(event=event)]
        
        # Compute transitions order
        transitions = self._sort_transitions(transitions)

        # Should the step consume an event?
        event = None if transitions[0].event is None else event
        
        return self._create_steps(event, transitions)


    def _create_steps(self, event: Optional[Event],
                      transitions: Iterable[Transition]) -> List[MicroStep]:
        """
        Return a (possibly empty) list of micro steps. Each micro step corresponds to the process of a transition
        matching given event.

        :param event: the event to consider, if any
        :param transitions: the transitions that should be processed
        :return: a list of micro steps.
        """
        returned_steps = []
        for transition in transitions:
            # Internal transition
            if transition.target is None:
                returned_steps.append(MicroStep(event=event, transition=transition))
                continue

            lca = self._statechart.least_common_ancestor(transition.source, transition.target)
            from_ancestors = self._statechart.ancestors_for(transition.source)
            to_ancestors = self._statechart.ancestors_for(transition.target)

            # Exited states
            exited_states = []

            # last_before_lca is the "highest" ancestor of from_state that is a child of LCA
            last_before_lca = transition.source
            for state in from_ancestors:
                if state == lca:
                    break
                last_before_lca = state

            # Take all the descendants of this state and list the ones that are active
            for descendant in self._statechart.descendants_for(last_before_lca)[::-1]:  # Mind the reversed order!
                # Only leave states that are currently active
                if descendant in self._configuration:
                    exited_states.append(descendant)

            # Add last_before_lca as it is a child of LCA that must be exited
            if last_before_lca in self._configuration:
                exited_states.append(last_before_lca)

            # Entered states
            entered_states = [transition.target]
            for state in to_ancestors:
                if state == lca:
                    break
                entered_states.insert(0, state)

            returned_steps.append(MicroStep(event=event, transition=transition,
                                            entered_states=entered_states, exited_states=exited_states))

        return returned_steps

    def _create_stabilization_step(self, names: Iterable[str]) -> Optional[MicroStep]:
        """
        Return a stabilization step, ie. a step that lead to a more stable situation
        for the current statechart. Stabilization means:

         - Enter the initial state of a compound state with no active child
         - Enter the memory of a history state
         - Enter the children of an orthogonal state with no active child
         - Empty active configuration if root's child is a final state

        :param names: List of states to consider (usually, the active configuration)
        :return: A *MicroStep* instance or *None* if this statechart can not be more stabilized
        """
        # Check if we are in a set of "stable" states
        leaves_names = self._statechart.leaf_for(names)
        leaves = sorted([self._statechart.state_for(name) for name in leaves_names],
                        key=lambda s: (-self._statechart.depth_for(s.name), s.name))

        for leaf in leaves:
            if isinstance(leaf, FinalState) and self._statechart.parent_for(leaf.name) == self._statechart.root:
                return MicroStep(exited_states=[leaf.name, self._statechart.root])
            if isinstance(leaf, (ShallowHistoryState, DeepHistoryState)):
                states_to_enter = self._memory.get(leaf.name, [leaf.memory])
                states_to_enter.sort(key=lambda x: (self._statechart.depth_for(x), x))
                return MicroStep(entered_states=states_to_enter, exited_states=[leaf.name])
            elif isinstance(leaf, OrthogonalState) and self._statechart.children_for(leaf.name):
                return MicroStep(entered_states=sorted(self._statechart.children_for(leaf.name)))
            elif isinstance(leaf, CompoundState) and leaf.initial:
                return MicroStep(entered_states=[leaf.initial])

        return None

    def _apply_step(self, step: MicroStep) -> MicroStep:
        """
        Apply given *MicroStep* on this statechart

        :param step: *MicroStep* instance
        :return: a new MicroStep, completed with sent events
        """
        entered_states = list(map(self._statechart.state_for, step.entered_states))
        exited_states = list(map(self._statechart.state_for, step.exited_states))

        active_configuration = set(self._configuration)  # Copy

        sent_events = []  # type: List[Event]

        # Exit states
        for state in exited_states:
            # Execute exit action
            sent_events.extend(self._evaluator.execute_on_exit(state))

            # Deal with history
            if isinstance(state, CompoundState):
                # Look for an HistoryStateMixin among its children
                for child_name in self._statechart.children_for(state.name):
                    child = self._statechart.state_for(child_name)
                    if isinstance(child, DeepHistoryState):
                        # This MUST contain at least one element!
                        active = active_configuration.intersection(self._statechart.descendants_for(state.name))
                        assert len(active) >= 1
                        self._memory[child.name] = list(active)
                    elif isinstance(child, ShallowHistoryState):
                        # This MUST contain exactly one element!
                        active = active_configuration.intersection(self.statechart.children_for(state.name))
                        assert len(active) == 1
                        self._memory[child.name] = list(active)

            # Remove state from active configuration
            self._configuration.remove(state.name)

            # Postconditions
            self._evaluate_contract_conditions(state, 'postconditions', step)

            # Notify properties
            self._notify_properties('state exited', state=state.name)

        # Execute transition
        if step.transition:
            # Preconditions and invariants
            self._evaluate_contract_conditions(step.transition, 'preconditions', step)
            self._evaluate_contract_conditions(step.transition, 'invariants', step)

            sent_events.extend(self._evaluator.execute_action(step.transition, step.event))

            # Postconditions and invariants
            self._evaluate_contract_conditions(step.transition, 'postconditions', step)
            self._evaluate_contract_conditions(step.transition, 'invariants', step)

            # Notify properties
            self._notify_properties(
                'transition processed',
                source=step.transition.source,
                target=step.transition.target,
                event=step.event
            )

        # Enter states
        for state in entered_states:
            # Preconditions
            self._evaluate_contract_conditions(state, 'preconditions', step)

            # Execute entry action
            sent_events.extend(self._evaluator.execute_on_entry(state))

            # Update configuration
            self._configuration.add(state.name)

            # Notify properties
            self._notify_properties('state entered', state=state.name)

        # Send events
        for event in sent_events:
            self._raise_event(event)

        return MicroStep(event=step.event, transition=step.transition,
                         entered_states=step.entered_states, exited_states=step.exited_states,
                         sent_events=sent_events)

    def _stabilize(self) -> List[MicroStep]:
        """
        Compute, apply and return stabilization steps.

        :return: A list of applied  *MicroStep* instances,
        """
        # Stabilization
        steps = []
        step = self._create_stabilization_step(self._configuration)
        while step is not None:
            steps.append(self._apply_step(step))
            step = self._create_stabilization_step(self._configuration)
        return steps

    def _evaluate_contract_conditions(self, obj: Union[Transition, StateMixin],
                                      cond_type: str,
                                      step: Union[MacroStep, MicroStep]=None) -> None:
        """
        Evaluate the conditions for given object.

        :param obj: object with preconditions, postconditions or invariants
        :param cond_type: either "preconditions", "postconditions" or "invariants"
        :param step: step in which the check occurs.
        :raises ContractError: if a condition fails and *ignore_contract* is False.
        """
        if self._ignore_contract:
            return

        exception_klass = cast(Callable[..., Exception], {'preconditions': PreconditionError,
                                                          'postconditions': PostconditionError,
                                                          'invariants': InvariantError}[cond_type])

        unsatisfied_conditions = getattr(self._evaluator, 'evaluate_' + cond_type)(obj, getattr(step, 'event', None))

        for condition in unsatisfied_conditions:
            raise exception_klass(configuration=self.configuration, step=step, obj=obj,
                                  assertion=condition, context=self.context)

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self._statechart)



import pytest

import pickle

from collections import Counter

from sismic.exceptions import ExecutionError, NonDeterminismError, ConflictingTransitionsError
from sismic.code import DummyEvaluator
from sismic.interpreter import Interpreter, Event, InternalEvent
from sismic.helpers import coverage_from_trace, log_trace, run_in_background
from sismic.model import Transition, MacroStep, MicroStep, MetaEvent


class TestInterpreterWithSimple:
    @pytest.fixture()
    def interpreter(self, simple_statechart):
        interpreter = Interpreter(simple_statechart, evaluator_klass=DummyEvaluator)

        # Stabilization
        interpreter.execute_once()

        return interpreter

    def test_init(self, interpreter):
        assert interpreter.configuration == ['root', 's1']
        assert not interpreter.final

    def test_time(self, interpreter):
        assert interpreter.time == 0

        interpreter.time += 10
        assert interpreter.time == 10

        interpreter.time = 20
        assert interpreter.time == 20

        with pytest.raises(ExecutionError):
            interpreter.time = 10

    def test_queue(self, interpreter):
        interpreter.queue(Event('e1'))
        assert interpreter._select_event(consume=True) == Event('e1')

        interpreter.queue(InternalEvent('e2'))
        assert interpreter._select_event(consume=True) == InternalEvent('e2')

        # Internal events are handled as external events, and thus have no priority
        interpreter.queue(Event('external'))
        interpreter.queue(InternalEvent('internal'))
        assert interpreter._select_event(consume=True) == Event('external')
        assert interpreter._select_event(consume=True) == InternalEvent('internal')

        interpreter.queue('e3')
        assert interpreter._select_event(consume=True) == Event('e3')

        interpreter.queue('e4').queue('e5')
        assert interpreter._select_event(consume=True) == Event('e4')
        assert interpreter._select_event(consume=True) == Event('e5')

        interpreter.queue('e6', 'e7', Event('e8'))
        assert interpreter._select_event(consume=True) == Event('e6')
        assert interpreter._select_event(consume=True) == Event('e7')
        assert interpreter._select_event(consume=True) == Event('e8')

    def test_simple_configuration(self, interpreter):
        assert interpreter.execute_once() is None  # Should do nothing!

        interpreter.queue(Event('goto s2')).execute_once()

        assert interpreter.configuration == ['root', 's2']

        interpreter.execute_once()
        assert interpreter.configuration == ['root', 's3']

    def test_simple_entered(self, interpreter):
        interpreter.queue(Event('goto s2'))
        assert interpreter.execute_once().entered_states == ['s2']

        interpreter.queue(Event('goto final'))
        assert interpreter.execute_once().entered_states == ['s3']
        assert interpreter.execute_once().entered_states == ['final']
        assert interpreter.configuration == []
        assert interpreter.final

    def test_simple_final(self, interpreter):
        interpreter.queue(Event('goto s2')).queue(Event('goto final')).execute()
        assert interpreter.final


class TestInterpreterWithInternal:
    @pytest.fixture
    def interpreter(self, internal_statechart):
        interpreter = Interpreter(internal_statechart)

        # Stabilization
        interpreter.execute_once()

        return interpreter

    def test_sent(self, interpreter):
        step = interpreter.execute_once()

        assert step.event.name == 'next'

    def test_internal_before_external(self, interpreter):
        assert interpreter.queue(Event('not_next')).execute_once().event.name == 'next'

        step = interpreter.execute_once()
        assert step.event is None
        assert step.entered_states == ['s2']

        assert interpreter.execute_once().event.name == 'not_next'

    def test_active_guard(self, interpreter):
        interpreter.execute()
        assert interpreter.final


class TestInterpreterWithFinal:
    @pytest.fixture()
    def interpreter(self, final_statechart):
        return Interpreter(final_statechart, evaluator_klass=DummyEvaluator)

    def test_not_final(self, interpreter):
        assert 'root' not in interpreter.configuration
        assert not interpreter.final

    def test_root_when_not_final(self, interpreter):
        interpreter.execute()
        assert interpreter.configuration == ['root', 'a']
        assert not interpreter.final

    def test_when_root_child_is_final(self, interpreter):
        interpreter.queue('root-final').execute()
        assert interpreter.configuration == []
        assert interpreter.final

    def test_when_root_child_is_not_final(self, interpreter):
        interpreter.queue('b').execute()
        assert interpreter.configuration == ['root', 'b']
        assert not interpreter.final

    def test_when_nested_child_is_final(self, interpreter):
        interpreter.queue('c').execute()
        assert interpreter.configuration == ['root', 'c', 'nested-final']
        assert not interpreter.final

    def test_with_parallel_states(self, interpreter):
        interpreter.queue('parallel').execute()
        assert interpreter.configuration == ['root', 'parallel', 'p1', 'p2', 'p1a', 'p2a']
        assert not interpreter.final

        interpreter.queue('p1-final').execute()
        assert interpreter.configuration == ['root', 'parallel', 'p1', 'p2', 'p1-final', 'p2a']
        assert not interpreter.final

        interpreter.queue('p2-final').execute()
        assert interpreter.configuration == ['root', 'parallel', 'p1', 'p2', 'p1-final', 'p2-final']
        assert not interpreter.final


class TestInterpreterWithNonDeterministic:
    @pytest.fixture()
    def interpreter(self, nondeterministic_statechart):
        interpreter = Interpreter(nondeterministic_statechart, evaluator_klass=DummyEvaluator)

        # Stabilization
        interpreter.execute_once()

        return interpreter

    def test_nondeterminism(self, interpreter):
        with pytest.raises(NonDeterminismError):
            interpreter.execute_once()


class TestInterpreterWithHistory:
    @pytest.fixture()
    def interpreter(self, history_statechart):
        interpreter = Interpreter(history_statechart, evaluator_klass=DummyEvaluator)

        # Stabilization
        interpreter.execute_once()

        return interpreter

    def test_memory(self, interpreter):
        interpreter.queue(Event('next')).execute_once()
        assert interpreter.configuration == ['root', 'loop', 's2']

        step = interpreter.queue(Event('pause')).execute_once()
        assert step.exited_states == ['s2', 'loop']
        assert interpreter.configuration == ['root', 'pause']

    def test_resume_memory(self, interpreter):
        interpreter.queue(
            Event('next'),
            Event('pause'),
            Event('continue')
        )
        last_step = interpreter.execute()[-1]

        assert last_step.entered_states == ['loop', 'loop.H', 's2']
        assert last_step.exited_states == ['pause', 'loop.H']
        assert interpreter.configuration == ['root', 'loop', 's2']

    def test_after_memory(self, interpreter):
        interpreter.queue(
            Event('next'),
            Event('pause'),
            Event('continue'),
            Event('next'),
            Event('next')
        ).execute()

        assert interpreter.configuration == ['root', 'loop', 's1']

        interpreter.queue(Event('pause'), Event('stop')).execute()
        assert interpreter.final


class TestInterpreterWithDeephistory:
    @pytest.fixture()
    def interpreter(self, deep_history_statechart):
        interpreter = Interpreter(deep_history_statechart, evaluator_klass=DummyEvaluator)

        # Stabilization
        interpreter.execute_once()

        return interpreter

    def test_deep_memory(self, interpreter):
        interpreter.queue(
            Event('next1'),
            Event('next2')
        ).execute()
        assert set(interpreter.configuration) == {'active', 'concurrent_processes', 'process_1',
                                                  'process_2', 'root', 's12', 's22'}

        interpreter.queue(Event('error1')).execute()
        assert interpreter.configuration == ['root', 'pause']
        assert set(interpreter._memory['active.H*']) == {'concurrent_processes', 'process_1', 'process_2',
                                                         's12', 's22'}

        interpreter.queue(Event('continue')).execute()
        assert set(interpreter.configuration) == {'active', 'concurrent_processes', 'process_1',
                                                  'process_2', 'root', 's12', 's22'}

    def test_entered_order(self, interpreter):
        interpreter.queue(
            Event('next1'),
            Event('next2'),
            Event('pause')
        )
        last_step = interpreter.execute()[-1]

        assert last_step.entered_states == ['pause']
        assert interpreter.configuration == ['root', 'pause']

        step = interpreter.queue(Event('continue')).execute_once()

        assert step.entered_states.index('active') < step.entered_states.index('active.H*')
        assert step.entered_states.index('active.H*') < step.entered_states.index('concurrent_processes')
        assert step.entered_states.index('concurrent_processes') < step.entered_states.index('process_1')
        assert step.entered_states.index('concurrent_processes') < step.entered_states.index('process_2')
        assert step.entered_states.index('process_1') < step.entered_states.index('s12')
        assert step.entered_states.index('process_2') < step.entered_states.index('s22')

        interpreter.queue(Event('next1')).queue(Event('next2')).execute()
        assert 's13' in interpreter.configuration and 's23' in interpreter.configuration
        assert not interpreter.final

    def test_exited_order(self, interpreter):
        interpreter.queue(
            Event('next1'),
            Event('next2'),
            Event('pause')
        )
        last_step = interpreter.execute()[-1]

        assert last_step.exited_states == ['s12', 's22', 'process_1', 'process_2', 'concurrent_processes', 'active']
        assert interpreter.configuration == ['root', 'pause']

        step = interpreter.queue(Event('continue')).execute_once()
        assert step.exited_states == ['pause', 'active.H*']

        interpreter.queue(Event('next1'), Event('next2')).execute()
        assert 's13' in interpreter.configuration and 's23' in interpreter.configuration
        assert not interpreter.final


class TestInterpreterWithInfinite:
    @pytest.fixture()
    def interpreter(self, infinite_statechart):
        interpreter = Interpreter(infinite_statechart)

        # Stabilization
        interpreter.execute_once()

        return interpreter

    def test_three_steps(self, interpreter):
        assert interpreter.configuration == ['root', 's1']
        interpreter.execute_once()
        assert interpreter.configuration == ['root', 's2']
        interpreter.execute_once()
        assert interpreter.configuration == ['root', 's1']
        interpreter.execute_once()
        assert interpreter.configuration == ['root', 's2']

        # x is incremented in s1.on_entry
        assert interpreter.context['x'] == 2

    def test_auto_three_steps(self, interpreter):
        interpreter.execute(max_steps=3)
        assert interpreter.configuration == ['root', 's2']

        # x is incremented in s1.on_entry
        assert interpreter.context['x'] == 2

    def test_auto_stop(self, interpreter):
        interpreter.execute()

        assert interpreter.final
        assert interpreter.context['x'] == 100


class TestInterpreterWithParallel:
    @pytest.fixture()
    def interpreter(self, parallel_statechart):
        interpreter = Interpreter(parallel_statechart, evaluator_klass=DummyEvaluator)

        # Stabilization
        interpreter.execute_once()

        return interpreter

    def test_concurrent_transitions(self, interpreter):
        step = interpreter.queue(Event('nextA')).execute_once()

        assert interpreter.configuration == ['root', 's1', 'p1', 'p2', 'a1', 'a2']
        assert step.exited_states.index('initial1') < step.exited_states.index('initial2')
        assert step.entered_states.index('a1') < step.entered_states.index('a2')

    def test_concurrent_transitions_nested_target(self, interpreter):
        interpreter.queue(Event('nextA'), Event('reset1')).execute()

        assert interpreter.configuration == ['root', 's1', 'p1', 'p2', 'a2', 'initial1']

    def test_unnested_transitions(self, interpreter):
        interpreter.queue(Event('nextA'), Event('nextA')).execute()
        assert interpreter.configuration == ['root', 's1', 'p1', 'p2', 'a2', 'initial1']

    def test_unnested_transitions_2(self, interpreter):
        interpreter.queue(Event('nextA'), Event('nextB')).execute()

        assert interpreter.configuration == ['root', 's1', 'p1', 'p2', 'b1', 'b2']

    def test_conflicting_transitions(self, interpreter):
        interpreter.queue(Event('nextA'), Event('nextB'), Event('conflict1'))
        interpreter.execute_once()
        interpreter.execute_once()

        with pytest.raises(ConflictingTransitionsError):
            interpreter.execute_once()

    def test_conflicting_transitions_2(self, interpreter):
        interpreter.queue(Event('nextA'), Event('nextB'), Event('conflict2'))
        interpreter.execute_once()
        interpreter.execute_once()

        with pytest.raises(ConflictingTransitionsError):
            interpreter.execute_once()


class TestInterpreterWithNestedParallel:
    common_states = ['root', 's1', 'p1', 'p2', 'r1', 'r2', 'r3', 'r4']

    @pytest.fixture()
    def interpreter(self, nested_parallel_statechart):
        interpreter = Interpreter(nested_parallel_statechart, evaluator_klass=DummyEvaluator)

        # Stabilization
        interpreter.execute_once()

        return interpreter

    def test_initial(self, interpreter):
        assert interpreter.configuration == self.common_states + ['i1', 'i2', 'i3', 'i4']

    def test_parallel_order(self, interpreter):
        step = interpreter.queue(Event('next')).execute_once()

        assert interpreter.configuration == self.common_states + ['j1', 'j2', 'j3', 'j4']
        assert step.exited_states == ['i1', 'i2', 'i3', 'i4']
        assert step.entered_states == ['j1', 'j2', 'j3', 'j4']
        assert [t.source for t in step.transitions] == ['i1', 'i2', 'i3', 'i4']

    def test_partial_parallel_order(self, interpreter):
        interpreter.queue(Event('next'), Event('click'))
        interpreter.execute_once()
        step = interpreter.execute_once()

        assert interpreter.configuration == self.common_states + ['j1', 'j3', 'k2', 'k4']
        assert step.exited_states == ['j2', 'j4']
        assert step.entered_states == ['k2', 'k4']
        assert [t.source for t in step.transitions] == ['j2', 'j4']

    def test_partial_unnested_transition(self, interpreter):
        interpreter.queue(Event('next'), Event('reset'))
        interpreter.execute_once()
        step = interpreter.execute_once()

        assert interpreter.configuration == self.common_states + ['i1', 'i2', 'i3', 'i4']
        assert step.exited_states.index('r2') < step.exited_states.index('r4')
        assert step.exited_states.index('p1') < step.exited_states.index('p2')
        assert step.exited_states.index('r2') < step.exited_states.index('p1')
        assert step.exited_states.index('r4') < step.exited_states.index('p2')
        assert step.entered_states.index('p1') < step.entered_states.index('p2')
        assert step.entered_states.index('p1') < step.entered_states.index('r1')
        assert step.entered_states.index('p1') < step.entered_states.index('r2')
        assert step.entered_states.index('r1') < step.entered_states.index('p2')
        assert step.entered_states.index('r2') < step.entered_states.index('p2')
        assert step.entered_states.index('p2') < step.entered_states.index('r3')
        assert step.entered_states.index('p2') < step.entered_states.index('r4')
        assert [t.source for t in step.transitions] == ['r2', 'r4']

    def test_name_order(self, interpreter):
        interpreter.queue(
            Event('next'),
            Event('click'),
            Event('next'),
            Event('next')
        )
        interpreter.execute_once()
        interpreter.execute_once()
        interpreter.execute_once()

        assert interpreter.configuration == self.common_states + ['k1', 'k3', 'x', 'y']

        step = interpreter.execute_once()
        assert step.exited_states.index('k1') < step.exited_states.index('k3')
        assert step.exited_states.index('k3') < step.exited_states.index('x')
        assert step.exited_states.index('x') < step.exited_states.index('y')
        assert interpreter.configuration == self.common_states + ['k1', 'x', 'y', 'z']
        assert step.entered_states.index('k1') < step.entered_states.index('z')
        assert step.entered_states.index('z') < step.entered_states.index('x')
        assert step.entered_states.index('x') < step.entered_states.index('y')

        assert [t.source for t in step.transitions] == ['k1', 'k3', 'x', 'y']

        step = interpreter.queue(Event('next')).execute_once()

        assert step.exited_states.index('k1') < step.exited_states.index('x')
        assert step.exited_states.index('x') < step.exited_states.index('y')
        assert step.exited_states.index('y') < step.exited_states.index('z')
        assert interpreter.configuration == self.common_states + ['k1', 'x', 'y', 'z']
        assert step.entered_states.index('k1') < step.entered_states.index('x')
        assert step.entered_states.index('x') < step.entered_states.index('y')
        assert step.entered_states.index('y') < step.entered_states.index('z')

        assert [t.source for t in step.transitions] == ['k1', 'x', 'y', 'z']


class TestLogTrace:
    @pytest.fixture(autouse=True)
    def setup(self, elevator):
        self.steps = log_trace(elevator)

    def test_empty_trace(self, elevator):
        assert self.steps == []

    def test_nonempty_trace(self, elevator):
        elevator.queue(Event('floorSelected', floor=4)).execute()
        assert len(self.steps) > 0

    def test_log_content(self, elevator):
        steps = elevator.queue(Event('floorSelected', floor=4)).execute()
        assert steps == self.steps


def test_run_in_background(elevator):
    from time import sleep

    task = run_in_background(elevator, 0.001)
    elevator.queue(Event('floorSelected', floor=4))

    sleep(0.01)

    task.stop()

    assert elevator.context['current'] == 4


class TestCoverageFromTrace:
    def test_empty_trace(self):
        assert coverage_from_trace([]) == {'entered states': Counter(), 'exited states': Counter(), 'processed transitions': Counter()}

    def test_single_step(self):
        trace = [MacroStep(0, steps=[
            MicroStep(entered_states=['a', 'b', 'c'], exited_states=['b'], transition=Transition('x')),
            MicroStep(entered_states=['a', 'b'], exited_states=['c'], transition=Transition('x')),
            MicroStep(entered_states=['a']),
            MicroStep(entered_states=[])
        ])]

        expected = {
            'entered states': Counter(a=3, b=2, c=1),
            'exited states': Counter(b=1, c=1),
            'processed transitions': Counter({Transition('x'): 2})
        }

        assert coverage_from_trace(trace) == expected

    def test_multiple_steps(self):
        trace = [MacroStep(0, steps=[
            MicroStep(entered_states=['a', 'b', 'c'], exited_states=['b'], transition=Transition('x')),
            MicroStep(entered_states=['a', 'b'], exited_states=['c'],  transition=Transition('x')),
            MicroStep(entered_states=['a']),
            MicroStep(entered_states=[])
        ])]

        trace.extend(trace)

        expected = {
            'entered states': Counter(a=6, b=4, c=2),
            'exited states': Counter(b=2, c=2),
            'processed transitions': Counter({Transition('x'): 4})
        }

        assert coverage_from_trace(trace) == expected


class TestInterpreterBinding:
    @pytest.fixture()
    def interpreter(self, simple_statechart):
        i1 = Interpreter(simple_statechart, evaluator_klass=DummyEvaluator)
        i2 = Interpreter(simple_statechart, evaluator_klass=DummyEvaluator)

        # Stabilization
        i1.execute_once(), i2.execute_once()

        return i1, i2

    def test_bind(self, interpreter):
        i1, i2 = interpreter

        i1.bind(i2)
        assert i2.queue in i1._bound

        i1._raise_event(InternalEvent('test'))
        assert i1._internal_events.pop() == Event('test')
        assert i2._external_events.pop() == Event('test')

    def test_bind_callable(self, interpreter):
        i1, i2 = interpreter

        i1.bind(i2.queue)
        assert i2.queue in i1._bound

        i1._raise_event(InternalEvent('test'))

        assert i1._internal_events.pop() == Event('test')
        assert i2._external_events.pop() == Event('test')

    def test_metaevent(self, interpreter):
        i1, i2 = interpreter

        i1.bind(i2.queue)

        i1._raise_event(MetaEvent('test'))

        assert len(i1._internal_events) == 0
        assert len(i2._external_events) == 0

    def test_event(self, interpreter):
        i1, i2 = interpreter

        i1.bind(i2.queue)

        with pytest.raises(ValueError):
            i1._raise_event(Event('test'))

        assert len(i1._internal_events) == 0
        assert len(i2._external_events) == 0


def test_interpreter_is_serialisable(microwave):
    microwave.queue(
        'door_opened',
        'item_placed',
        'door_closed',
        'timer_inc',
        'timer_inc',
        'cooking_start'
        'timer_tick',
    ).execute()

    dumped = pickle.dumps(microwave)
    n_microwave = pickle.loads(dumped)

    assert microwave.configuration == n_microwave.configuration
    assert microwave.context == n_microwave.context



import pytest

from sismic import code
from sismic.code.python import FrozenContext
from sismic.exceptions import CodeEvaluationError
from sismic.interpreter import Event, InternalEvent, MetaEvent


def test_dummy_evaluator(mocker):
        interpreter = mocker.MagicMock(name='interpreter')
        evaluator = code.DummyEvaluator(interpreter=interpreter)

        assert evaluator._evaluate_code('blablabla') is True
        assert evaluator._evaluate_code('False') is True
        assert evaluator.context == {}

        assert evaluator._execute_code('blablabla') == []
        assert evaluator.context == {}


def test_frozen_context():
    context = {'a': 1, 'b': 2}

    freeze = FrozenContext(context)
    assert len(freeze) == 2
    assert freeze.a == 1
    assert freeze.b == 2

    context['a'] = 2
    assert freeze.a == 1


class TestPythonEvaluator:
    @pytest.fixture
    def evaluator(self, mocker):
        context = {
            'x': 1,
            'y': 2,
            'z': 3
        }

        interpreter = mocker.MagicMock(name='Interpreter')
        interpreter.time = 0
        interpreter.queue = mocker.MagicMock(return_value=None)
        interpreter.statechart = mocker.MagicMock()
        interpreter.configuration = []

        return code.PythonEvaluator(interpreter, initial_context=context)

    @pytest.fixture
    def interpreter(self, evaluator):
        return evaluator._interpreter

    def test_condition(self, evaluator):
        assert evaluator._evaluate_code('True')
        assert not evaluator._evaluate_code('False')
        assert evaluator._evaluate_code('1 == 1')
        assert evaluator._evaluate_code('x == 1')
        with pytest.raises(CodeEvaluationError):
            evaluator._evaluate_code('a')

    def test_condition_on_event(self, evaluator):
        assert evaluator._evaluate_code('event.data[\'a\'] == 1', additional_context={'event': Event('test', a=1)})
        assert evaluator._evaluate_code('event.name == \'test\'', additional_context={'event': Event('test')})

    def test_execution(self, evaluator):
        evaluator._execute_code('a = 1')
        assert evaluator.context['a'] == 1
        evaluator._execute_code('x = 2')
        assert evaluator.context['x'] == 2

    def test_invalid_condition(self, evaluator):
        with pytest.raises(CodeEvaluationError):
            evaluator._evaluate_code('x.y')

    def test_invalid_action(self, evaluator):
        with pytest.raises(CodeEvaluationError):
            evaluator._execute_code('x = x.y')

    def test_send(self, evaluator):
        events = evaluator._execute_code('send("hello")')
        assert events == [InternalEvent('hello')]

    def test_notify(self, evaluator):
        events = evaluator._execute_code('notify("hello", x=1, y="world")')
        assert events == [MetaEvent('hello', x=1, y='world')]

    def test_no_event_raised_by_preamble(self, interpreter, evaluator):
        interpreter.statechart.preamble = 'send("test")'
        with pytest.raises(CodeEvaluationError):
            evaluator.execute_statechart(interpreter.statechart)

    def test_add_variable_in_context(self, evaluator):
        evaluator._execute_code('a = 1\nassert a == 1', additional_context=evaluator.context)
        assert evaluator._evaluate_code('a == 1', additional_context={'a': 1})

    @pytest.mark.skip('http://stackoverflow.com/questions/32894942/listcomp-unable-to-access-locals-defined-in-code-called-by-exec-if-nested-in-fun')
    def test_access_outer_scope(self, evaluator):
        evaluator._execute_code('d = [x for x in range(10) if x!=a]', additional_context={'a': 1})



import pytest

from sismic.exceptions import (InvariantError, PostconditionError,
                               PreconditionError)
from sismic.interpreter import Interpreter, Event
from sismic.model import StateMixin, Transition


def test_no_error(elevator):
    elevator.queue(Event('floorSelected', floor=4))
    elevator.execute()

    assert not elevator.final


def test_state_precondition(elevator):
    elevator.statechart.state_for('movingUp').preconditions.append('False')
    elevator.queue(Event('floorSelected', floor=4))

    with pytest.raises(PreconditionError) as e:
        elevator.execute()

    assert isinstance(e.value.obj, StateMixin)


def test_state_postcondition(elevator):
    elevator.statechart.state_for('movingUp').postconditions.append('False')
    elevator.queue(Event('floorSelected', floor=4))

    with pytest.raises(PostconditionError) as e:
        elevator.execute()

    assert isinstance(e.value.obj, StateMixin)


def test_state_invariant(elevator):
    elevator.statechart.state_for('movingUp').invariants.append('False')
    elevator.queue(Event('floorSelected', floor=4))

    with pytest.raises(InvariantError) as e:
        elevator.execute()

    assert isinstance(e.value.obj, StateMixin)


def test_transition_precondition(elevator):
    transitions = elevator.statechart.transitions_from('floorSelecting')
    transitions[0].preconditions.append('False')
    elevator.queue(Event('floorSelected', floor=4))

    with pytest.raises(PreconditionError) as e:
        elevator.execute()

    assert isinstance(e.value.obj, Transition)


def test_transition_postcondition(elevator):
    transitions = elevator.statechart.transitions_from('floorSelecting')
    transitions[0].postconditions.append('False')
    elevator.queue(Event('floorSelected', floor=4))

    with pytest.raises(PostconditionError) as e:
        elevator.execute()

    assert isinstance(e.value.obj, Transition)


def test_do_not_raise(elevator):
    elevator = Interpreter(elevator.statechart, ignore_contract=True)
    transitions = elevator.statechart.transitions_from('floorSelecting')
    transitions[0].postconditions.append('False')

    elevator.queue(Event('floorSelected', floor=4)).execute()



import tkinter as tk
from functools import partial

from sismic.interpreter import Interpreter, Event
from sismic.io import import_from_yaml

# The two following lines are NOT needed in a typical environment.
# These lines make sismic available in our testing environment
import sys
sys.path.append('../../..')




####################################################

# Create a tiny GUI
class MicrowaveApplication(tk.Frame):
    def __init__(self, master=None):
        super().__init__(master)

        # Initialize widgets
        self.create_widgets()

        # Create a Stopwatch interpreter
        with open('microwave.yaml') as f:
            statechart = import_from_yaml(f)
        self.interpreter = Interpreter(statechart)

        # Bind interpreter events to the GUI
        self.interpreter.bind(self.event_handler)

        self.execute()

    def execute(self):
        self.interpreter.execute()

        # Update the widget that contains the list of active states.
        self.w_states['text'] = '\n'.join(self.interpreter.configuration)

        self.w_timer['text'] = 'M.timer: %d' % self.interpreter.context.get('controller.timer', 'undefined')
        self.w_power['text'] = 'M.power: %d' % self.interpreter.context.get('controller.power', 'undefined')

    def create_widgets(self):
        self.pack(fill=tk.BOTH)

        # MAIN frame containing all others
        left_frame = tk.Frame(self)
        left_frame.pack(side=tk.LEFT, fill=tk.BOTH)

        # INPUT frame
        input_frame = tk.LabelFrame(left_frame, text='INPUT BUTTONS')
        input_frame.pack(side=tk.LEFT, fill=tk.BOTH, padx=(8, 8))

        self.w_power_inc = tk.Button(input_frame, text='power +',
                                     command=partial(self.send_event, event_name='input_power_inc'))
        self.w_power_dec = tk.Button(input_frame, text='power -',
                                     command=partial(self.send_event, event_name='input_power_dec'))
        self.w_power_reset = tk.Button(input_frame, text='power reset',
                                       command=partial(self.send_event, event_name='input_power_reset'))

        self.w_power_inc.pack(side=tk.TOP, fill=tk.X)
        self.w_power_dec.pack(side=tk.TOP, fill=tk.X)
        self.w_power_reset.pack(side=tk.TOP, fill=tk.X)

        self.w_timer_inc = tk.Button(input_frame, text='timer +',
                                     command=partial(self.send_event, event_name='input_timer_inc'))
        self.w_timer_dec = tk.Button(input_frame, text='timer -',
                                     command=partial(self.send_event, event_name='input_timer_dec'))
        self.w_timer_reset = tk.Button(input_frame, text='timer 0',
                                       command=partial(self.send_event, event_name='input_timer_reset'))

        self.w_timer_inc.pack(side=tk.TOP, fill=tk.X, pady=(8, 0))  # leave some space before first button
        self.w_timer_dec.pack(side=tk.TOP, fill=tk.X)
        self.w_timer_reset.pack(side=tk.TOP, fill=tk.X)

        self.w_cooking_start = tk.Button(input_frame, text='start',
                                         command=partial(self.send_event, event_name='input_cooking_start'))
        self.w_cooking_stop = tk.Button(input_frame, text='stop',
                                        command=partial(self.send_event, event_name='input_cooking_stop'))

        self.w_cooking_start.pack(side=tk.TOP, fill=tk.X, pady=(8, 0))  # leave some space before first button
        self.w_cooking_stop.pack(side=tk.TOP, fill=tk.X)

        # SENSORS frame
        sensors_frame = tk.LabelFrame(left_frame, text='SENSORS')
        sensors_frame.pack(side=tk.LEFT, fill=tk.BOTH, padx=(8, 8))

        # SENSORS > Clock frame
        clock_frame = tk.LabelFrame(sensors_frame, text='Clock')
        clock_frame.pack(side=tk.TOP, fill=tk.BOTH)
        self.w_clock = tk.Label(clock_frame)

        self.w_tick = tk.Button(clock_frame, text='tick', command=partial(self.send_event, event_name='timer_tick'))
        self.w_tick.pack(side=tk.TOP, fill=tk.X)

        # SENSORS > WeightSensor frame
        weight_frame = tk.LabelFrame(sensors_frame, text='WeightSensor')
        weight_frame.pack(side=tk.TOP, fill=tk.BOTH, pady=(8, 0))
        self.w_weight = tk.Label(weight_frame)

        self.w_item_placed = tk.Button(weight_frame, text='place item',
                                       command=partial(self.send_event, event_name='item_placed'))
        self.w_item_removed = tk.Button(weight_frame, text='remove item',
                                        command=partial(self.send_event, event_name='item_removed'))

        self.w_item_placed.pack(side=tk.TOP, fill=tk.X)
        self.w_item_removed.pack(side=tk.TOP, fill=tk.X)

        # SENSORS > Door frame
        door_frame = tk.LabelFrame(sensors_frame, text='Door')
        door_frame.pack(side=tk.TOP, fill=tk.BOTH, pady=(8, 0))
        self.w_door = tk.Label(door_frame)

        self.w_door_opened = tk.Button(door_frame, text='open door',
                                       command=partial(self.send_event, event_name='door_opened'))
        self.w_door_closed = tk.Button(door_frame, text='close door',
                                       command=partial(self.send_event, event_name='door_closed'))

        self.w_door_opened.pack(side=tk.TOP, fill=tk.X)
        self.w_door_closed.pack(side=tk.TOP, fill=tk.X)

        # ACTUATORS frame
        right_frame = tk.LabelFrame(self, text='ACTUATORS')
        right_frame.pack(side=tk.LEFT, fill=tk.BOTH, padx=(8, 8))

        # Display component
        display_frame = tk.LabelFrame(right_frame, text='Display')
        display_frame.pack(side=tk.TOP, fill=tk.BOTH)
        self.w_display = tk.Label(display_frame)
        self.w_display.pack(side=tk.TOP)

        # Lamp component
        lamp_frame = tk.LabelFrame(right_frame, text='Lamp')
        lamp_frame.pack(side=tk.TOP, fill=tk.BOTH)
        self.w_lamp = tk.Label(lamp_frame)
        self.w_lamp.pack(side=tk.TOP)

        # Heating component
        heating_frame = tk.LabelFrame(right_frame, text='Heating')
        heating_frame.pack(side=tk.TOP, fill=tk.BOTH)
        self.w_heating_power = tk.Label(heating_frame)
        self.w_heating_status = tk.Label(heating_frame)
        self.w_heating_power.pack(side=tk.TOP)
        self.w_heating_status.pack(side=tk.TOP)

        # Beeper component
        beep_frame = tk.LabelFrame(right_frame, text='Beeper')
        beep_frame.pack(side=tk.TOP, fill=tk.BOTH)
        self.w_beep = tk.Label(beep_frame)
        self.w_beep.pack(side=tk.TOP)

        # Turntable component
        turntable_frame = tk.LabelFrame(right_frame, text='Turntable')
        turntable_frame.pack(side=tk.TOP, fill=tk.BOTH)
        self.w_turntable = tk.Label(turntable_frame)
        self.w_turntable.pack(side=tk.TOP)

        # Oven controller statechart component
        controller_frame = tk.LabelFrame(self, text='Oven Controller')
        controller_frame.pack(side=tk.LEFT, fill=tk.BOTH, padx=(8, 8))
        # Microwave Controller Statechart status
        # statechart_frame = tk.LabelFrame(right_frame2, text='Controller')
        # statechart_frame.pack(side=tk.TOP, fill=tk.BOTH)

        states_frame = tk.LabelFrame(controller_frame, text='States')
        states_frame.pack(side=tk.TOP, fill=tk.BOTH)
        self.w_states = tk.Label(states_frame)
        self.w_states.pack(side=tk.TOP, fill=tk.X, pady=(8, 0))

        variables_frame = tk.LabelFrame(controller_frame, text='Variables')
        variables_frame.pack(side=tk.TOP, fill=tk.BOTH, pady=(8, 0))
        self.w_timer = tk.Label(variables_frame)
        self.w_timer.pack(side=tk.BOTTOM, fill=tk.X)

        self.w_power = tk.Label(variables_frame)
        self.w_power.pack(side=tk.BOTTOM, fill=tk.X)

    def event_handler(self, event):
        name = event.name

        if name == 'lamp_switch_on':
            self.w_lamp['text'] = 'on'
        elif name == 'lamp_switch_off':
            self.w_lamp['text'] = 'off'
        elif name == 'display_set':
            self.w_display['text'] = event.text
        elif name == 'display_clear':
            self.w_display['text'] = ''
        elif name == 'heating_set_power':
            self.w_heating_power['text'] = event.power
        elif name == 'heating_on':
            self.w_heating_status['text'] = 'on'
        elif name == 'heating_off':
            self.w_heating_status['text'] = 'off'
        elif name == 'beep':
            self.w_beep['text'] = event.number
        elif name == 'turntable_start':
            self.w_turntable['text'] = 'on'
        elif name == 'turntable_stop':
            self.w_turntable['text'] = 'off'
        else:
            raise ValueError('Unknown event %s' % event)

    def send_event(self, event_name):
        self.interpreter.queue(Event(event_name))
        self.execute()

    def _quit(self):
        self.master.destroy()


if __name__ == '__main__':
    # Create GUI
    root = tk.Tk()
    root.wm_title('Microwave Oven')
    app = MicrowaveApplication(master=root)

    app.mainloop()



from sismic.helpers import log_trace


def before_scenario(context, scenario):
    # Create interpreter
    statechart = context.config.userdata.get('statechart')
    interpreter_klass = context.config.userdata.get('interpreter_klass')
    context.interpreter = interpreter_klass(statechart)

    # Log trace
    context.trace = log_trace(context.interpreter)
    context._monitoring = False
    context.monitored_trace = None

    # Bind property statecharts
    for property_statechart in context.config.userdata.get('property_statecharts'):
        context.interpreter.bind_property_statechart(interpreter_klass(property_statechart))


def before_step(context, step):
    # "Then" steps must at least follow one "when" step
    if step.step_type == 'then':
        # Stop monitoring
        context._monitoring = False

        if context.monitored_trace is None:
            raise ValueError('Scenario must at least contain one "when" step before any "then" step.')


def after_step(context, step):
    # "Given" triggers execution
    if step.step_type == 'given':
        context.interpreter.execute()

    # "When" triggers monitored execution
    if step.step_type == 'when':
        macrosteps = context.interpreter.execute()

        if not context._monitoring:
            context._monitoring = True
            context.monitored_trace = []

        context.monitored_trace.extend(macrosteps)

    # Hook to enable debugging
    if step.step_type == 'then' and step.status == 'failed' and context.config.userdata.get('debug_on_error'):
        try:
            import ipdb as pdb
        except ImportError:
            import pdb

        print('--------------------------------------------------------------')
        print('Dropping into (i)pdb.', end='\n\n')
        print('Variable context holds the current execution context of Behave')
        print('You can access the interpreter using context.interpreter, the')
        print('trace using context.trace and the monitored trace using')
        print('context.monitored_trace.')
        print('--------------------------------------------------------------')

        pdb.post_mortem(step.exc_traceback)